"""
Defines Building-related constants.
"""
BUILD_NAMES = ['FullServiceRestaurant',    #1
                'Hospital',    #2
                'LargeHotel',    #3
                'LargeOffice',    #4
                'MediumOffice',    #5
                'MidriseApartment',    #6
                'OutPatient',    #7
                'PrimarySchool',    #8
                'QuickServiceRestaurant',    #9
                'SecondarySchool',    #10
                'SmallHotel',    #11
                'SmallOffice',    #12
                'Stand-aloneRetail',    #13
                'StripMall',    #14
                'SuperMarket',    #15
                'Warehouse']     #16

CLIMATE_NAMES = {'1A':'1A_USA_FL_MIAMI', '2A':'2A_USA_TX_HOUSTON', '2B': '2B_USA_AZ_PHOENIX', '3A': '3A_USA_GA_ATLANTA',
                    '3Bcoast': '3B_USA_CA_LOS_ANGELES', '3B': '3B_USA_NV_LAS_VEGAS', '3C': '3C_USA_CA_SAN_FRANCISCO',
                    '4A': '4A_USA_MD_BALTIMORE', '4B': '4B_USA_NM_ALBUQUERQUE', '4C': '4C_USA_WA_SEATTLE',
                    '5A': '5A_USA_IL_CHICAGO-OHARE', '5B': '5B_USA_CO_BOULDER', '6A': '6A_USA_MN_MINNEAPOLIS',
                    '6B': '6B_USA_MT_HELENA', '7A': '7A_USA_MN_DULUTH', '8A': '8A_USA_AK_FAIRBANKS'}

IDF_FIELDS = {
    'Building': ['north_axis','terrain','load_tol','temp_tol','solar_distribution','max_warm_up','min_warm_up',],
    'Site:Location': ['latitude','longitude','timezone','elevation'],
    'RunPeriod':['m1','d1','m2','d2','day','weather_special','weather_daylight','weekend_holiday','rain_inidcators','snow_indicators','repeat'],
    'RunPeriodControl:SpecialDays':['start','duration','type'],
    'Material': ['roughness','thickness','conductivity','density','specheat','thermal_absorptance','solar_absorptance','visible_absorptance',],
    'Material:NoMass': ['roughness','thermal_resistance','thermal_absorptance','solar_absorptance','visible_absorptance',],
    'WindowMaterial:SimpleGlazingSystem': ['u_factor','solar_heat_gain','visible_transmittance',],
    'WindowMaterial:Glazing': ['optical_data_type','spectral_data_set','thickness','solar_transmittance','reflectance_front','reflectance_back','transmittance_visible','reflectance_visible_front','reflectance_visible_back','transmittance_infrared','emittance_front','emittance_back','thermal_conductivity','dirt_correction','solar_diffusing',],
    'WindowMaterial:Gas': ['gas','gap_thickness',],
    'BuildingSurface:Detailed':['surf_type','construct','zone','boundary','object'],
    'InternalMass':['construct','zone'],
    'FenestrationSurface:Detailed':['surf_type','construct','surf_name','object','view_factor','shading','frame_name','multiplier'],
    'WindowProperty:FrameAndDivider':['width','projection_out','projection_in','conductance','conductance_ratio_edge_center','solar_absoprtance','visible_absorptance','emissivity','divider_type','divider_width','horizontal_dividers','vertical_dividers','divider_projection_out','divider_projection_in','divider_conductance','divider_conductance_ratio','divider_solar_absoprtnace','divider_visible_absorptance','divider_emissivity'],
    'Zone':['relative_north','x_origin','y_origin','z_origin','zone_type','multiplier','ceil_height','volume','floor_area','inside_convection','outside_convection','incl_in_total_area'],
    'Pump:VariableSpeed':['inlet','outlet','design_flow','design_head','design_power','motor_eff','fluid_ineff_frac','c1','c2','c3','c4','design_min_flow','control_type','schedule'],
    'Pump:ConstantSpeed':['inlet','outlet','design_flow','design_head','design_power','motor_eff','fluid_ineff_frac','control_type','schedule'],
    'Chiller:Electric:EIR':['capacity','COP_ref','chilled_water_temperature_ref','condenser_water_temperature_ref','chilled_water_flow_ref','condenser_water_flow_ref','capacity_temperature_curve','electric_in_out_temperature_curve','electric_in_out_part_load_curve','min_part_load','max_part_load','opt_part_load','min_unload_ratio','chilled_water_inlet','chilled_water_outlet','condenser_inlet','condenser_outlet','condenser_type','fan_power_ratio','frac_power_2_condenser','min_chilled_water_temperature','flow_mode','heat_recovery_flow','heat_recovery_inlet','heat_recovery_outlet','sizing_factor'],
    'Chiller:Electric:ReformulatedEIR':['capacity','COP_ref','chilled_water_temperature_ref','condenser_water_temperature_ref','chilled_water_flow_ref','condenser_water_flow_ref','capacity_temperature_curve','electric_in_out_temperature_curve','electric_in_out_curve_type','electric_in_out_part_load_curve','min_part_load','max_part_load','opt_part_load','min_unload_ratio','chilled_water_inlet','chilled_water_outlet','condenser_inlet','condenser_outlet','frac_power_2_condenser','min_chilled_water_temperature','flow_mode','heat_recovery_flow','heat_recovery_inlet','heat_recovery_outlet','sizing_factor'],
    'Boiler:HotWater':['fuel','capacity','efficiency','eff_temperature_curve','normalized_eff_curve','design_water_temperature','design_flow_rate','min_part_load','max_part_load','optimal_part_load','water_inlet','water_outlet','max_water_temperature','flow_mode','parasitic_electric_load','sizing_factor'],
    'Humidifier:Steam:Electric':['schedule','flow_capacity','max_power','fan_power','standby_power'],
    'AirLoopHVAC:UnitaryHeatCool':['schedule','inlet','outlet','fan_schedule','max_temperature','cooling_air_flow','heating_air_flow','no_load_air_flow','thermostat','fan_type','fan_name','fan_location','heat_coil_type','heat_coil','cool_coil_type','cool_coil','dehumid_control','reheat_type','reheat_name'],
    'CoolingTower:SingleSpeed':['water_inlet','water_outlet','water_flow','air_flow','fan_power','UA','free_convect_flow','free_convect_factor','free_convect_UA','free_convect_UA_factor','performance_method','heat_reject_capacity','nominal_capacity','free_convect_capacity','free_convect_size_factor','des_inlet_dry_bulb','des_inlet_wet_bulb','des_approach_temp','des_range_temp','basin_heater_capacity','basin_heater_setpoint','basin_heater_schedule','evap_loss_mode','evap_loss_factor','drift_loss','blowdown_mode','blowdown_ratio','blowdown_makeup_schedule','supply_water_tank','outdoor_air_inlet','capacity_control','number_of_cells','cell_control','cell_min_water_frac','cell_max_water_frac','sizing_factor'],
    'WaterHeater:Mixed':['tank_volume','temperature_schedule','deadband_temperature_dif','max_temperature','heater_control_type','heater_maximum_capacity','heater_minimum_capacity','heater_ignition_minimum_flow','heater_ignition_delay','heater_fuel','heater_efficiency','part_load_curve','off_cycle_fuel_use','off_cycle_fuel_type','off_cycle_frac2tank','on_cycle_fuel_use','on_cycle_fuel_type','on_cycle_frac2tank','ambient_temp_indicator','ambient_temp_schedule','ambient_temp_zone','ambient_temp_outdoor_node','off_cycle_loss2ambient','off_cycle_loss2zone','on_cycle_loss2ambient','on_cycle_loss2zone','peak_use_flow','use_flow_schedule','cold_water_supply_temp_schedule','use_side_inlet','use_side_outlet','use_side_effectiveness','source_side_inlet','source_side_outlet','source_side_effectiveness','use_side_design_flow','source_side_design_flow','indirect_water_heating_recovery_time'],
    'Controller:WaterCoil':['control_variable','action','actuator_variable','sensor_node','actuator_node','tolerance','max_actuated_flow','min_actuated_flow'],
    'SetpointManager:Scheduled':['control_variable','schedule','node'],
    'SetpointManager:MixedAir':['control_variable','reference_setpoint_node','fan_inlet_node','fan_outlet_node','node'],
    'SetpointManager:SingleZone:Reheat':['control_variable','min_temp','max_temp','zone','zone_node','zone_inlet_node','node'],
    'SetpointManager:SingleZone:Humidity:Minimum':['node','control_zone_node'],
    'SetpointManager:SingleZone:Humidity:Maximum':['node','control_zone_node'],
    'SetpointManager:FollowOutdoorAirTemperature':['control_variable','ref_temp_type','offset_temperature','max_temperature','min_temperature','node'],
    'SetpointManager:OutdoorAirReset':['control_variable','setpoint_at_low_temp','low_temp','setpoint_at_high_temp','high_temp','node'],
    'ZoneControl:Thermostat':['zone','control_type_schedule','control_type','control_name'],
    'ZoneControl:Humidistat':['zone','humidify_schedule','dehumidify_schedule'],
    'ThermostatSetpoint:SingleHeating':['heating'],
    'ThermostatSetpoint:SingleCooling':['cooling'],
    'ThermostatSetpoint:SingleHeatingOrCooling':['heating'],
    'ThermostatSetpoint:DualSetpoint':['heating','cooling'],
    'DesignSpecification:OutdoorAir':['method','flow_person','flow_area','flow_zone'],
    'Sizing:Zone':['cooling_T_method','cooling_T_des','cooling_dT_des','heating_T_method','heating_T_des','heating_dT_des','cooling_w_des','heating_w_des','outdoor_air_object','heating_sizing','cooling_sizing','cooling_flow_des_method','cooling_flow_des','cooling_min_flow_per_area','cooling_min_flow','cooling_min_frac','heating_flow_des_method','heating_flow_des','heating_min_flow_per_area','heating_min_flow','heating_min_frac'],
    'ZoneHVAC:UnitHeater':['schedule','inlet','outlet','fan_type','fan_name','max_air_flow','heat_coil_type','heat_coil','fan_schedule','fan_without_heating','max_hot_water','min_hot_water','tolerance','availability'],
    'ZoneHVAC:PackagedTerminalAirConditioner':['schedule','inlet','outlet','outdoor_air_type','outdoor_air','cooling_air_flow','heating_air_flow','no_load_air_flow','cooling_outdoor_air_flow','heating_outdoor_air_flow','no_load_outdoor_air_flow','fan_type','fan_name','heat_coil_type','heat_coil','cool_coil_type','cool_coil','fan_placement','fan_schedule'],
    'ZoneHVAC:FourPipeFanCoil':['schedule','capacity_control','max_air_flow','low_flow_ratio','medium_flow_ratio','outdoor_air_flow','outdoor_air_schedule','inlet','outlet','outdoor_air_mixer_type','outdoor_air_mixer','fan_type','fan_name','cool_coil_type','cool_coil','max_cold_water','min_cold_water','cooling_tolerance','heat_coil_type','heat_coil','max_hot_water','min_hot_water','heating_tolerance'],
    'Fan:ConstantVolume':['schedule','fan_efficiency','pressure_rise','flow_rate','motor_efficiency','motor_frac','inlet','outlet','end_use'],
    'Fan:ZoneExhaust':['schedule','fan_efficiency','pressure_rise','flow_rate','inlet','outlet','end_use'],
    'Fan:VariableVolume':['schedule','fan_efficiency','pressure_rise','flow_rate','method','min_flow_frac','min_flow_rate','motor_efficiency','motor_frac','a1','a2','a3','a4','a5','inlet','outlet','end_use'],
    'Fan:OnOff':['schedule','fan_efficiency','pressure_rise','flow_rate','motor_efficiency','motor_frac','inlet','outlet','power_speed_curve','efficiency_speed_curve','end_use'],
    'Curve:Quadratic':['a0','a1','a2','min_x','max_x'],
    'Curve:Cubic':['a0','a1','a2','a3','min_x','max_x'],
    'Curve:Biquadratic':['a0','a1','a2','b1','b2','ab','min_x','max_x','min_y','max_y'],
    'Curve:Bicubic':['a0','a1','a2','b1','b2','ab','a3','b3','aab','abb','min_x','max_x','min_y','max_y'],
    'Curve:ChillerPartLoadWithLift':['a0','a1','a2','b1','b2','ab','a3','b3','aab','abb','aabb','cb3','min_x','max_x','min_y','max_y'],
    'Coil:Cooling:DX:TwoSpeed':['schedule','rated_capacity','rated_sensible_heat_ratio','rated_COP','rated_air_flow','static_air_pressure','inlet','outlet','capacity_v_temperature_curve','capacity_v_flow_curve','energy_input_v_temperature_curve','energy_input_v_flow_curve','part_load_curve','rated_capacity2','rated_sensible_heat_ratio2','rated_COP2','rated_air_flow2','capacity_v_temperature_curve2','energy_input_v_temperature_curve2','condenser_inlet'],
    'Coil:Cooling:DX:SingleSpeed':['schedule','rated_capacity','rated_sensible_heat_ratio','rated_COP','rated_air_flow','rated_fan_power_per_volume','inlet','outlet','capacity_v_temperature_curve','capacity_v_flow_curve','energy_input_v_temperature_curve','energy_input_v_flow_curve','part_load_curve','min_ambient_temp',],
    'Coil:Cooling:Water':['schedule','rated_water_flow','rated_air_flow','water_inlet_temperature','air_inlet_temperature','air_outlet_temperature','air_inlet_humididty','air_outlet_humidity','inlet_water','outlet_water','inlet','outlet','analysis_type','heat_exchanger_configuration'],
    'Coil:Cooling:Water:DetailedGeometry':['schedule','rated_water_flow','tube_area_outside','tube_area_inside','fin_area','min_airflow_area','coil_depth','fin_diameter','fin_thickness','tube_id','tube_od','tube_conductivity','fin_conductivity','fin_spacing','tube_spacing','tube_rows','tubes_per_row','inlet_water','outlet_water','inlet','outlet'],
    'Coil:Heating:Fuel':['schedule','fuel','efficiency','capacity','inlet','outlet','temperature_setpoint'],
    'Coil:Heating:Water':['schedule','UA','rated_water_flow','inlet_water','outlet_water','inlet','outlet','input_method','capacity','water_inlet_temperature','air_inlet_temperature','water_outlet_temperature','air_outlet_temperature','air_water_convection_ratio'],
    'Coil:Heating:Electric':['schedule','efficiency','capacity','inlet','outlet','setpoint_node'],
    'Site:WaterMainsTemperature':['schedule','outdoor_average','max_temp_difference'],
    'EnvironmentalImpactFactors':['district_cool_cop','steam_efficiency','carbon_factor_n2o','carbon_factor_ch4','carbon_factor_co2'],
    'CoilSystem:Cooling:DX':['schedule','inlet','outlet','sensor','coil_type','coil_name'],
    'ZoneInfiltration:DesignFlowRate':['zone_name','schedule','method','nominal_set','per_area','per_ext_area','changes_per_hr','c1','t1','v1','v2'],
    'ZoneMixing':['receiving_zone','schedule','mix_type','nominal_set','per_area','per_ext_area','changes_per_area','source_zone'],
    'People':['zone_name','schedule','method','people','people_per_area','area_per_person','radiant','sensible_frac','activity','co2_rate','comfort_warnings','radiant_calc','surface_name','work_eff','insulation_method','insulation_calk_schedule','insulation_schedule','velocity_schedule','comfort_model'],
    'Lights':['zone_name','schedule','method','light_level','watt_area','watt_person','return_frac','radiant','visible','replacable','end_use','calc_from_plenum'],
    'ElectricEquipment':['zone_name','schedule','method','equip_level','watt_area','watt_person','latent','radiant','lost','end_use'],
    'GasEquipment':['zone_name','schedule','method','equip_level','watt_area','watt_person','latent','radiant','lost','end_use'],
    'Exterior:Lights':['schedule','nominal','control','category'],
    'Exterior:FuelEquipment':['fuel','schedule','nominal','category'],
    'Refrigeration:Case':['schedule','zone','rated_ambient_T','rated_ambient_RH','capacity_per_length','latent_heat_ratio','runtime_fraction','length','temperature','latent_credit_curve_type','latent_credit_curve_name','fan_power_per_length','operating_fan_power_per_length','standard_lighting_per_unit_length','installed_standard_lighting_per_unit_length','light_schedule','light_to_case','anti_sweat_heater_per_length','minimum_anti_sweat_per_length','anti_sweat_control','humidity_at_zero_percent','height','anti_sweat_heat_to_case','defrost_power_per_length','defrost_type','defrost_schedule','defrost_drip_down_schedule','defrost_curve_type','defrost_curve_name','return_air_frac','restock_schedule','case_credit_fraction_schedule'],
    'Refrigeration:CompressorRack':['heat_reject_location','design_COP','COP_curve','fan_power','fan_power_temperature_curve','condensor_type','water_inlet','water_outlet','water_loop_type','water_condensor_temperature_schedule','water_flow_rate','water_max_flow','water_max_outlet_T','water_min_inlet_T','evap_condenser_schedule','evap_condenser_effectiveness','evap_condenser_air_flow','basin_heater_capacity','basin_heater_setpoint','water_pump_power','water_supply_tank','condenser_air_inlet','end_use','case_name',],
    'Refrigeration:CaseAndWalkInList':['case'],
    'WaterUse:Equipment':['end_use_category','peak_flow','flow_schedule','target_temperature_schedule','hot_supply_temperature_schedule','cold_supply_temperature_schedule','zone','sensible_frac_schedule','latent_frac_schedule'],
    'WaterUse:Connections':['inlet','outlet','supply_tank','reclaimation_tank','hot_schedule','cold_schedule','drain_HX_type','drain_HX_destination','drain_HX_UA','equip_name'],
    'AirLoopHVAC':['controller','manager','supply_flow','branch_list','connector_list','supply_inlet','demand_outlet','demand_inlet','supply_outlet'],
    'BranchList':['branches'],
    'Sizing:Parameters':['cooling_sizing_factor','averaging_windows'],
    'SizingPeriod:DesignDay':['month','day','day_type','drybulb_temp','drybulb_range','db_range_modifier_type','db_range_modifier_schedule','humidity_type','wetbulb_temp','humidity_schedule','humidity_at_max','enthalpy_at_max','wetbulb_range','pressure','windspeed','winddir','rain_indicate','snow_indicate','daylight_savings_indicate','beam_solar_day_indicate','diffuse_solar_day_schedule','clear_sky_depth_beam','clear_sky_depth_diffuse','clearness'],
    'Sizing:System':['sizing_load_type','outdoor_flow','heating_max_flow','T_preheat','w_preheat','T_precool','w_precool','Tsupply_c','Tsupply_h','zone_sum_type','cool_all_outdoor_air','heat_all_outdoor_air','w_supply_c','w_supply_h','cooling_supply_flow_rate_method','cooling_supply_flow_rate','cooling_supply_flow_rate_area','cool_frac_autosized','cooling_supply_flow_rate_per_capacity','heating_supply_flow_rate','heating_supply_flow_rate_area','heat_frac_autosized','heating_supply_flow_rate_per_capacity','outdoor_air_method','cool_capacity_method','cool_capacity','cool_capacity_area','frac_autosize_cooling','heat_capacity_method','heat_capacity','heat_capacity_area','frac_autosize_heating','control_method'],
    'Branch':['pressure_curve','component_type','component_name','component_inlet','component_outlet'],
    'AirLoopHVAC:SupplyPath':['inlet','component_type','component_name'],
    'AirLoopHVAC:ReturnPath':['outlet','component_type','component_name'],
    'AirLoopHVAC:ZoneSplitter':['inlet','outlets'],
    'AirLoopHVAC:ZoneMixer':['outlet','inlets'],
    'AirLoopHVAC:ReturnPlenum':['zone','zone_node','outlet','induced_air','inlets'],
    'ZoneHVAC:EquipmentConnections':['hvac_equipment','inlet','exhaust','node_name','return_'],
    'ZoneHVAC:EquipmentList':['e_type','e_name','cool_sequence','heat_sequence'],
    'AirTerminal:SingleDuct:Uncontrolled':['schedule','inlet','max_flow'],
    'AirTerminal:SingleDuct:VAV:Reheat':['schedule','outlet_damper','inlet','max_flow','min_method','min_frac','min_flow','flow_frac_schedule','reheat_coil_type','reheat_coil_name','max_flow_water','min_flow_water','outlet','tolerance','damper_action','max_reheat_flow_per_area','max_reheat_frac'],
    'ZoneHVAC:AirDistributionUnit':['outlet','terminal_type','terminal'],
    'AirLoopHVAC:OutdoorAirSystem':['controller_list','equipment_list','manager_list'],
    'AirLoopHVAC:ControllerList':['controller_type','controller_name'],
    'Controller:OutdoorAir':['relief_outlet','return_air','mixed_air','actuator_node','min_flow','max_flow','economizer_control','economizer_action','economizer_max_T','economizer_max_h','economizer_max_dp','enthalpy_limit_curve','economizer_min_T','lockout','min_limit_type','min_air_schedule','min_frac_schedule','max_frac_schedule','mechanical_controller_name'],
    'AvailabilityManagerAssignmentList':['manager_type','manager_name'],
    'AvailabilityManager:NightCycle':['schedule','fan_schedule','control_type','thermostat_tolerance','runtime_control_type','runtime'],
    'AvailabilityManager:Scheduled':['schedule'],
    'PlantLoop':['fluid','user_fluid','operation_scheme','setpoint_node','max_temperature','min_temperature','max_flow_rate','min_flow_rate','loop_volume','supply_inlet','supply_outlet','supply_branch_list','supply_connector_list','demand_inlet','demand_outlet','demand_branch_list','demand_connector_list','load_scheme'],
    'CondenserLoop':['fluid','user_fluid','operation_scheme','setpoint_node','max_temperature','min_temperature','max_flow_rate','min_flow_rate','loop_volume','supply_inlet','supply_outlet','supply_branch_list','supply_connector_list','demand_inlet','demand_outlet','demand_branch_list','demand_connector_list','load_scheme'],
    'Connector:Mixer':['outlet','inlets'],
    'Connector:Splitter':['inlet','outlets'],
} 

EIO_FIELDS = {}
EIO_FIELDS['AirTerminal:SingleDuct:Uncontrolled'] = {'location':['hvac','terminals','name'],
                                                    'Design Size Maximum Air Flow Rate [m3/s]':['max_flow'],
                                                    'User-Specified Maximum Air Flow Rate [m3/s]':['max_flow']}
EIO_FIELDS['AirTerminal:SingleDuct:VAV:Reheat'] ={'location':['hvac','terminals','name'],
                                                   'Design Size Maximum Air Flow Rate [m3/s]':['max_flow'],
                                                   'Design Size Constant Minimum Air Flow Fraction':['min_frac'],
                                                   'Design Size Minimum Air Flow Rate [m3/s]':['min_flow'],
                                                   'Design Size Maximum Flow per Zone Floor Area during Reheat [m3/s-m2]':['max_reheat_flow_per_area'],
                                                   'Design Size Maximum Flow Fraction during Reheat []':['max_reheat_frac'],
                                                   'Design Size Maximum Reheat Water Flow Rate [m3/s]':['max_flow_water'],
                                                   'User-Specified Constant Minimum Air Flow Fraction':['min_frac']}

EIO_FIELDS['Coil:Heating:Water'] = {'location':['coils_heating'],
                                    'Design Size Rated Capacity [W]':['design_capacity'],
                                    'Design Size Maximum Water Flow Rate [m3/s]':['rated_water_flow'],
                                    'Design Size U-Factor Times Area Value [W/K]':['UA'],
                                    'Nominal Total Capacity {W}':['capacity']}
EIO_FIELDS['Coil:Heating:Fuel'] = {'location':['coils_heating'],
                                   'Design Size Nominal Capacity [W]':['capacity']}

EIO_FIELDS['Coil:Heating:Electric'] = {'location':['coils_heating'],
                                       'Design Size Nominal Capacity [W]':['capacity']}
EIO_FIELDS['Coil:Cooling:Water'] = {'location':['coils_cooling'],
                                    'Design Size Design Coil Load [W]':['design_capacity'],
                                    'Design Size Design Water Flow Rate [m3/s]':['rated_water_flow'],
                                    'Design Size Design Air Flow Rate [m3/s]':['rated_air_flow'],
                                    'User-Specified Design Air Flow Rate [m3/s]':['rated_air_flow'],
                                    'Design Size Design Inlet Air Temperature [C]':['air_inlet_temperature'],
                                    'Design Size Design Inlet Water Temperature [C]':['water_inlet_temperature'],
                                    'Design Size Design Outlet Air Temperature [C]':['air_outlet_temperature'],
                                    'Design Size Design Inlet Air Humidity Ratio':['air_inlet_humididty'],
                                    'Design Size Design Outlet Air Humidity Ratio':['air_outlet_humididty'],
                                    'Nominal Total Capacity {W}':['capacity'],
                                    'Nominal Sensible Capacity {W}':['sensible_capacity'],
                                    'Nominal Latent Capacity {W}':['latent_capacity'],
                                    'Nominal Sensible Heat Ratio':['sensible_heat_ratio'],
                                    'Nominal Coil UA Value {W/C}':['UA'],
                                    'Nominal Coil Surface Area {m2}':['area']}
EIO_FIELDS['Coil:Cooling:Water:DetailedGeometry'] = {'location':['coils_cooling'],
                                    'Design Size Design Coil Load [W]':['design_capacity'],
                                    'Design Size Maximum Water Flow Rate [m3/s]':['rated_water_flow'],
                                    'Design Size Number of Tubes per Row':['tubes_per_row'],
                                    'Design Size Fin Diameter [m]':['fin_diameter'],
                                    'Design Size Minimum Airflow Area [m2]':['min_airflow_area'],
                                    'Design Size Fin Surface Area [m2]':['fin_area'],
                                    'Design Size Total Tube Inside Area [m2]':['tube_area_inside'],
                                    'Design Size Tube Outside Surface Area [m2]':['tube_area_outside'],
                                    'Design Size Coil Depth [m]':['coil_depth'],
                                    'Nominal Total Capacity {W}':['capacity'],
                                    'Nominal Sensible Capacity {W}':['sensible_capacity'],
                                    'Nominal Latent Capacity {W}':['latent_capacity'],
                                    'Nominal Sensible Heat Ratio':['sensible_heat_ratio']}
EIO_FIELDS['Coil:Cooling:DX:TwoSpeed'] = {'location':['coils_cooling'],
                                          'Design Size High Speed Rated Air Flow Rate [m3/s]':['rated_air_flow'],
                                          'Design Size High Speed Gross Rated Total Cooling Capacity [W]':['rated_capacity'],
                                          'Design Size High Speed Rated Sensible Heat Ratio':['rated_sensible_heat_ratio'],
                                          'Design Size Low Speed Rated Air Flow Rate [m3/s]':['rated_air_flow2'],
                                          'Design Size Low Speed Gross Rated Total Cooling Capacity [W]':['rated_capacity2'],
                                          'Design Size Low Speed Gross Rated Sensible Heat Ratio':['rated_sensible_heat_ratio2']}
EIO_FIELDS['Coil:Cooling:DX:SingleSpeed'] = {'location':['coils_cooling'],
                                            'Design Size Rated Air Flow Rate [m3/s]':['rated_air_flow'],
                                            'Design Size Rated Total Cooling Capacity [W]':['rated_capacity'],
                                            'Design Size Gross Rated Total Cooling Capacity [W]':['rated_capacity'],
                                            'Design Size Rated Sensible Heat Ratio':['rated_sensible_heat_ratio'],
                                            'Design Size Gross Rated Sensible Heat Ratio':['rated_sensible_heat_ratio']}
EIO_FIELDS['AirLoopHVAC'] = {'location':['hvac','loop','name'],
                            'Design Supply Air Flow Rate [m3/s]':['max_flow']}

EIO_FIELDS['AirLoopHVAC:UnitaryHeatCool'] = {'location':['unitary_heat_cool','name'],
                                            'Supply Air Flow Rate [m3/s]':['max_air_flow'],
                                            'Supply Air Flow Rate During Heating Operation [m3/s]':['heating_air_flow'],
                                            'Supply Air Flow Rate During Cooling Operation [m3/s]':['cooling_air_flow'],
                                            'Supply Air Flow Rate When No Cooling or Heating is Needed [m3/s]':['no_load_air_flow'],
                                            'Nominal Heating Capacity [W]':['capacity_heating'],
                                            'Nominal Cooling Capacity [W]':['capacity_cooling'],
                                            'Maximum Supply Air Temperature from Supplemental Heater [C]':['max_temperature'],
                                            'Fraction of Supply Air Flow That Goes Through the Controlling Zone':['frac_2_zone']}
EIO_FIELDS['Controller:OutdoorAir'] = {'location':['hvac','outdoor_air','control'],
                                       'Maximum Outdoor Air Flow Rate [m3/s]':['max_flow'],
                                       'Minimum Outdoor Air Flow Rate [m3/s]':['min_flow']}

EIO_FIELDS['Fan:ConstantVolume'] = {'location':['fans'],
                           'Design Size Maximum Flow Rate [m3/s]':['flow_rate']}
EIO_FIELDS['Fan:VariableVolume'] = {'location':['fans'],
                           'Design Size Maximum Flow Rate [m3/s]':['flow_rate']}
EIO_FIELDS['Fan:OnOff'] = {'location':['fans'],
                           'Design Size Maximum Flow Rate [m3/s]':['flow_rate']}
EIO_FIELDS['PlantLoop'] = {'location':['plant_loop','name'],
                               'Maximum Loop Flow Rate [m3/s]':['max_flow_rate'],
                               'Plant Loop Volume [m3]':['loop_volume'],
                               'Condenser Loop Volume [m3]':['loop_volume']}
EIO_FIELDS['CondenserLoop'] = {'location':['plant_loop','name'],
                               'Maximum Loop Flow Rate [m3/s]':['max_flow_rate'],
                               'Plant Loop Volume [m3]':['loop_volume'],
                               'Condenser Loop Volume [m3]':['loop_volume']}
EIO_FIELDS['Pump:VariableSpeed'] = {'location':['pump'],
                                    'Design Flow Rate [m3/s]':['design_flow'],
                                    'Design Power Consumption [W]':['design_power']}
EIO_FIELDS['Pump:ConstantSpeed'] = {'location':['pump'],
                                    'Design Flow Rate [m3/s]':['design_flow'],
                                    'Design Power Consumption [W]':['design_power']}
EIO_FIELDS['WaterHeater:Mixed'] = {'location':['water_heater'],
                                   'Use Side Design Flow Rate [m3/s]':['use_side_design_flow'],
                                   'Source Side Design Flow Rate [m3/s]':['source_side_design_flow']}
EIO_FIELDS['Controller:WaterCoil'] = {'location':['controller','name'],
                                      'Maximum Actuated Flow [m3/s]':['max_actuated_flow'],
                                      'Controller Convergence Tolerance':['tolerance']}
EIO_FIELDS['Chiller:Electric:ReformulatedEIR'] = {'location':['chiller'],
                                      'Design Size Reference Chilled Water Flow Rate [m3/s]':['chilled_water_flow_ref'],
                                      'Design Size Reference Capacity [W]':['capacity'],
                                      'Design Size Reference Condenser Water Flow Rate [m3/s]':['condenser_water_flow_ref'],
                                      'Design Size Reference Condenser Fluid Flow Rate  [m3/s]':['condenser_water_flow_ref']}
EIO_FIELDS['Chiller:Electric:EIR'] = {'location':['chiller'],
                                      'Design Size Reference Chilled Water Flow Rate [m3/s]':['chilled_water_flow_ref'],
                                      'Design Size Reference Capacity [W]':['capacity'],
                                      'Design Size Reference Condenser Water Flow Rate [m3/s]':['condenser_water_flow_ref'],
                                      'Design Size Reference Condenser Fluid Flow Rate  [m3/s]':['condenser_water_flow_ref']}
EIO_FIELDS['Boiler:HotWater'] = {'location':['boiler'],
                                 'Design Size Nominal Capacity [W]':['capacity'],
                                 'Design Size Design Water Flow Rate [m3/s]':['design_flow_rate']}
EIO_FIELDS['CoolingTower:SingleSpeed'] = {'location':['cooling_tower'],
                                                   'Design Water Flow Rate [m3/s]':['water_flow'],
                                                   'Fan Power at Design Air Flow Rate [W]':['fan_power'],
                                                   'Design Air Flow Rate [m3/s]':['air_flow'],
                                                   'U-Factor Times Area Value at Design Air Flow Rate [W/C]':['UA'],
                                                   'Free Convection Regime Air Flow Rate [m3/s]':['free_convect_flow'],                                                   
                                                   'Free Convection U-Factor Times Area Value [W/K]':['free_convect_UA']}
EIO_FIELDS['CoolingTower:TwoSpeed'] = {'location':['cooling_tower'],
                                                   'Design Water Flow Rate [m3/s]':['water_flow'],
                                                   'Fan Power at Design Air Flow Rate [W]':['fan_power'],
                                                   'Design Air Flow Rate [m3/s]':['air_flow'],
                                                   'U-Factor Times Area Value at Design Air Flow Rate [W/C]':['UA'],
                                                   'Free Convection Regime Air Flow Rate [m3/s]':['free_convect_flow'],                                                   
                                                   'Free Convection U-Factor Times Area Value [W/K]':['free_convect_UA']}
EIO_FIELDS['CoolingTower:VariableSpeed:Merkel'] = {'location':['cooling_tower'],
                                                   'Design Water Flow Rate [m3/s]':['water_flow'],
                                                   'Fan Power at Design Air Flow Rate [W]':['fan_power'],
                                                   'Design Air Flow Rate [m3/s]':['air_flow'],
                                                   'U-Factor Times Area Value at Design Air Flow Rate [W/C]':['UA'],
                                                   'Free Convection Regime Air Flow Rate [m3/s]':['free_convect_flow'],                                                   
                                                   'Free Convection U-Factor Times Area Value [W/K]':['free_convect_UA']}
EIO_FIELDS['Humidifier:Steam:Electric'] = {'location':['humidifiers'],
                                           'Design Size Nominal Capacity Volume [m3/s]':['flow_capacity'],
                                           'User-Specified Nominal Capacity Volume [m3/s]':['flow_capacity'],
                                           'Design Size Rated Power [W]':['max_power'],
                                           'User-Specified Rated Power [W]':['max_power']}
EIO_FIELDS['ZoneHVAC:UnitHeater'] = {'location':['unitary_sys','name'],
                                     'Design Size Maximum Supply Air Flow Rate [m3/s]':['max_air_flow']}
EIO_FIELDS['ZoneHVAC:FourPipeFanCoil'] = {'location':['unitary_sys','name'],
                                          'Design Size Maximum Supply Air Flow Rate [m3/s]':['max_air_flow'],
                                          'Design Size Maximum Hot Water Flow [m3/s]':['max_cold_water'],
                                          'Design Size Maximum Cold Water Flow [m3/s]':['max_hot_water']}
EIO_FIELDS['ZoneHVAC:PackagedTerminalAirConditioner'] = {'location':['unitary_sys','name'],
                                                         'Design Size Cooling Supply Air Flow Rate [m3/s]':['cooling_air_flow'],
                                                         'Design Size Heating Supply Air Flow Rate [m3/s]':['heating_air_flow'],
                                                         'Design Size No Load Supply Air Flow Rate [m3/s]':['no_load_air_flow'],
                                                         'Supply Air Flow Rate When No Cooling or Heating is Needed [m3/s]':['no_load_air_flow'],
                                                         'Design Size Outdoor Air Flow Rate During Cooling Operation [m3/s]':['cooling_outdoor_air_flow'],
                                                         'Design Size Outdoor Air Flow Rate During Heating Operation [m3/s]':['heating_outdoor_air_flow'],
                                                         'Design Size Outdoor Air Flow Rate When No Cooling or Heating is Needed [m3/s]':['no_load_outdoor_air_flow']}


import datetime 

from eagers.building.process.forecast_dr_capacity import forecast_dr_capacity
from eagers.basic.build_time_vector import build_time_vector
from eagers.basic.find_gen_node import find_gen_node
from eagers.setup.load_matrices import load_matrices
from eagers.setup.automatic_ic import automatic_ic
from eagers.simulate.renewable_output import renewable_output
from eagers.update.update_cost import update_cost
from eagers.forecasting.update_forecast import update_forecast
from eagers.forecasting.hydro_forecast import hydro_forecast
from eagers.setup.initialize_observer import initialize_observer
from eagers.solver.dispatch_loop import dispatch_loop


def water_year_forecast(gen,buildings,fluid_loop,observer,all_data_nodes,names,zones,pl,market,subnet,options,date,test_data):
    hydroforecast = False
    for i in range(len(gen)):
        if gen[i]['type']=='HydroStorage':
            hydroforecast = True
    if 'hydro_wy_timestamp' in observer and observer['hydro_wy_timestamp'] !=None:
        if observer['hydro_wy_timestamp'][-1]>=date[-1]:
            hydroforecast = False

    if hydroforecast:
        #first create the yearly dispatch data 
        # i.e. run Dispatch loop with updated information
        #these will be used for set points in the actual dispatch
        date_now = date[0]
        options['horizon'] = 364*24 #Yearly Horizon
        options['resolution'] = 7*24 #Week Resolution
        if date_now.month()<10 and date[-1].month()>=10:
            year = date_now.year()-1
            dy1 = datetime.datetime(date_now.year(),10,1,0,0,0)
            options['horizon'] = (53+round((date[-1]-dy1).days()/7+.5))*7*24 #ensure that water year forecast goes a week beyond the final date
        elif date_now.month()<10:
            year = date_now.year()-1
        else:
            year = date_now.year()
        d1 = datetime.datetime(year, 10, 1, 1, 0, 0) 
        date = build_time_vector(d1,options, to_timedelta=True)
        dt = [(date[t+1] - date[t]).seconds/3600 for t in range(len(date)-1)]
        b_names = [buildings[i].name for i in range(len(buildings))]
        op_mat_a = load_matrices(gen,b_names,fluid_loop,market,subnet,options,'A',dt) #build quadratic programming matrices for FitA
        op_mat_b = load_matrices(gen,b_names,fluid_loop,market,subnet,options,'B',dt) #build quadratic programming matrices for FitB
        one_step = load_matrices(gen,b_names,fluid_loop,market,subnet,options,'B',[]) #build quadratic programming matrices for single time step

        if observer['hydro_wy_timestamp'] == None:
            scale_cost = update_cost(date,gen,market)
            ic, data_t0, _ = automatic_ic(gen,buildings,fluid_loop,market,subnet,date[0],one_step,options,test_data,None,scale_cost) # set the initial conditions 
            wy_observer = initialize_observer(gen,subnet,fluid_loop,all_data_nodes, names, zones, pl,test_data,None,ic,50,date_now,data_t0,None)
        else:
            wy_observer = observer
        
        if date[-1]<=test_data['hydro']['timestamp'][-1]:
            forecast = update_forecast(options,date[1:],test_data,subnet)
            if 'weather' in forecast and 'dir_norm_irr' in forecast['weather']:
                forecast['renewable'] = [renewable_output(gen[i], date[1:], forecast['weather']['dir_norm_irr']) \
                                        if gen[i]['type'] in ['Renewable','Solar','Wind'] and gen[i]['enabled'] \
                                        else [] for i in range(len(gen))]
            forecast['building'],_,_,_ = forecast_dr_capacity(buildings,wy_observer,forecast['weather'],date)
            forecast = hydro_forecast(forecast,test_data,date[1:],subnet,[],options)
            solution,_ = dispatch_loop(gen,wy_observer,market,subnet,names,op_mat_a,op_mat_b,one_step,options,date,forecast,[])
            hydro_soc_init = [0 for n in range(len(subnet['hydro']['nodes']))]
            for n in range(len(subnet['hydro']['nodes'])):
                i = subnet['hydro']['equipment'][n]
                if gen[i]['type']=='HydroStorage':
                    hydro_soc_init[n] = observer.stor_state[i]
            solution['hydro_soc'] = [solution['hydro_soc'][n].insert(0,hydro_soc_init) for n in range(len(solution['hydro_soc']))]
            print('Water Year Forecast Completed for ' + str(year)+':',str(year+1))
            if observer.hydro_wy_timestamp==None:
                t = 0
                while solution['timestamp'][t+1]<date_now:
                    t+=1
                r = (solution['timestamp'][t+1]-date_now)/(solution['timestamp'][t+1]-solution['timestamp'][t])
                for i in range(len(gen)):
                    if gen[i]['type'] =='HydroStorage':
                        n = gen[i]['hydro']['subnet_node'] #dam #
                        observer['stor_state'][i] = r*solution['hydro_soc'][t][n] + (1-r)*solution['hydro_soc'][t+1][n]
            observer['hydro_wy_timestamp'] = solution['timestamp'] 
            observer['hydro_wy_soc'] = solution['hydro_soc']
    return observer



from eagers.building.basic.natural_convection import natural_convection

def exterior_convection(building,T_exterior_sur,Tair,wind_speed,leward,T_sky,method):
    sig = 5.67e-8 # W/m^2*K^4 Stephan-Boltzman const
    exterior = building.surfaces['exterior']
    n_ext = len(exterior['name'])
    F_grnd = [0.5*(1-exterior['cos_phi'][i]) for i in range(n_ext)]
    F_sky = [0.5*(1+exterior['cos_phi'][i]) for i in range(n_ext)]
    h_wind = [0 for i in range(n_ext)]
    beta = [F_sky[i]**.5 for i in range(n_ext)]
    h_n = natural_convection(exterior['normal'],[T_exterior_sur[i] - Tair[i] for i in range(n_ext)],None,method)
    
    h_ground = []
    h_air = []
    h_sky = []
    
    for i in range(n_ext):
        if exterior['boundary'][i] == 'ground':
            F_sky[i] = 0
            F_grnd[i] = 0
            h_wind[i] = 0
        else:
            if method=='DOE-2':
                if leward[i]:
                    h_wind[i] = h_n[i] + exterior['roughness_factor'][i]*((h_n[i]**2 + (3.55*wind_speed[i]**.617)**2)**.5-h_n[i])
                else:
                    h_wind[i] = h_n[i] + exterior['roughness_factor'][i]*((h_n[i]**2 + (3.26*wind_speed[i]**.89)**2)**.5-h_n[i])
            elif method=='Detailed' or method=='BLAST' or method=='TARP':
                if leward[i]:
                    h_wind[i] = h_n[i] + 2.537*.5*exterior['roughness_factor'][i]*(exterior['perimeter'][i]*wind_speed[i]/exterior['area'][i])**.5
                else:
                    h_wind[i] = h_n[i] + 2.537*exterior['roughness_factor'][i]*(exterior['perimeter'][i]*wind_speed[i]/exterior['area'][i])**.5    
        if exterior['boundary'][i] == 'ground':
            h_ground.append(building.ctf['ground_cond'][i])
            h_air.append(0)
        elif T_exterior_sur[i]==Tair[i]:
            h_ground.extend(exterior['thermal_absorptance'][i]*sig*F_grnd[i]*4*(Tair[i])**3) #avoid nan by using derivative
            h_air.extend(exterior['thermal_absorptance'][i]*sig*F_sky[i]*(1-beta[i])*4*(Tair[i])**3) #avoid nan by using derivative
        else:    
            h_ground.append(exterior['thermal_absorptance'][i]*sig*F_grnd[i]*((T_exterior_sur[i]+273)**4 - (Tair[i]+273)**4)/(T_exterior_sur[i] - Tair[i]))
            h_air.append(exterior['thermal_absorptance'][i]*sig*F_sky[i]*(1-beta[i])*((T_exterior_sur[i]+273)**4 - (Tair[i]+273)**4)/(T_exterior_sur[i] - Tair[i]))
        if T_exterior_sur[i]==T_sky:
            h_sky.append(exterior['thermal_absorptance'][i]*sig*F_sky[i]*beta[i]*4*(T_sky)**3) #avoid nan by using derivative
        else:
            h_sky.append(exterior['thermal_absorptance'][i]*sig*F_sky[i]*beta[i]*((T_exterior_sur[i]+273)**4 - (T_sky+273)**4)/(T_exterior_sur[i] - T_sky))
    h_s = [h_wind[i] + h_air[i] + h_ground[i] for i in range(n_ext)]
    return h_s,h_sky


from eagers.simulate.assign_disparity import assign_disparity
from eagers.simulate.building_response import building_response
from eagers.solver.dispatch_loop import dispatch_loop
from eagers.basic.all_demands import all_demands
from eagers.simulate.renewable_output import renewable_output
from eagers.basic.component_output import component_output
from eagers.update.update_cost import update_cost
from eagers.basic.marginal_cost import node_marginal


def plant_response(
        gen, building, fluid_loop, observer, market, subnet, names, actual_data,
        preload, options, forecast, date, solution):
    """Assigns where disparity between actual load and optimal dispatch
    goes. Building water loops absorb error in heating/cooling if there
    is a building model.
    """
    
    # 'control' mode runs real-time optimization that handles imbalance.
    if options['method'] =='control':
        # TODO: Reconstitute real-time control (Threshold and MPC).
        pass
    elif options['method'] =='dispatch':
        re_opt = False
        n_b = len(building)
        tol = dict(electrical = 1e-3, direct_current = 1e-3, transmission_1 = 1e-3, district_heat = 5e-2, district_cool = 5e-2, heating_2 = 5e-2, cooling_water = 5e-2, hydrogen = 1e-3, liq_hydrogen = 1e-3, hydro = 1e-3) 
        dt_now = (date[1] - date[0]).seconds/3600
        node_demand = all_demands(actual_data,subnet,[0])
        setpoint = {}
        if len(names['buildings'])>0:
            setpoint['temperature'] = [solution['building'][k]['temperature'][0] for k in names['buildings']]
        nn = list(node_demand.keys())
        # Sim all generators to determine actual heating, cooling, and waste heat to cooling loop
        for i in range(len(gen)):
            if gen[i]['type'] in ['Renewable','Solar','Wind']:
                renew = renewable_output(gen[i], [date[1]], actual_data['weather']['dir_norm_irr'])
                solution['dispatch'][i][1] = renew[0]
                solution['generator_state'][gen[i]['name']][0] = renew[0]
        gs_now = {}
        for k in solution['generator_state']:
            gs_now[k] = solution['generator_state'][k][0]
        scale_cost = update_cost(date[1:], gen,market)
        ##TODO add capacity (particularly capacity for response by buildings for use in assign disparity)
        marginal = node_marginal(gen,subnet,solution['dispatch'],scale_cost,solution['value_heat'])
        #sim fluid loops to assign thermal disparity to water loop
        for i in range(len(observer['fluid_loop_temperature'])):
            production = component_output(gen,subnet['cooling_water'],gs_now)
            observer['fluid_loop_temperature'][i] += dt_now/observer['fluid_loop_capacitance'][i]*production[i]
        #Assign disparity for heating/cooling to utility/storage/buildings or re-optimize if none are present
        for net in ['district_heat', 'district_cool']:
            if net in nn:
                production = component_output(gen,subnet[net],gs_now)
                request = building_request(names['buildings'],net,subnet[net],solution['building'],forecast['building'],[0])
                setpoint[net] = assign_disparity(net, names, gen, subnet[net],production,node_demand[net],request, solution, marginal, dt_now, tol[net], options['excess_heat'], re_opt)
                nn.remove(net)
        #Sim buildings 
        if n_b > 0:
            weather_now = {}
            for i in actual_data['weather']:
                weather_now[i] = actual_data['weather'][i][0]
            observer,net_electric = building_response(building,observer, weather_now, actual_data['building'], dt_now, setpoint,date[0])
            actual_data['building']['E0'][0] = net_electric
            actual_data['building']['C0'][0] = [solution['building'][k]['cooling'][0] for k in names['buildings']]
            actual_data['building']['H0'][0] = [solution['building'][k]['heating'][0] for k in names['buildings']]

        #rebalance other networks if necessary
        for net in nn:
            request =  None 
            if net == 'electrical' and 'buildings' in subnet[net] and any([len(n)>0 for n in subnet[net]['buildings']]):
                request = {}
                request['nominal'] = net_electric
            if not re_opt:
                production = component_output(gen,subnet[net],gs_now)
                assign_disparity(net, names, gen, subnet[net], production, node_demand[net], request, solution, marginal, dt_now, tol[net], False, re_opt)
        if re_opt:
            # Redo optimization due to disparity with actual data
            # and no slack bus, e.g. utility or storage.
            for k in forecast:
                if isinstance(forecast[k],dict) and 'demand' in forecast[k]:
                    forecast[k]['demand'][0] = actual_data[k]['demand'][0]
            if n_b > 0:
                forecast['building']['h_min'][0] = actual_data['building']['H0'][0]
                forecast['building']['c_min'][0] = actual_data['building']['C0'][0]
                forecast['building']['E0'][0] = actual_data['building']['E0'][0] 
                forecast['building']['C0'][0] = actual_data['building']['C0'][0]
                forecast['building']['H0'][0] = actual_data['building']['H0'][0]

            solution, _ = dispatch_loop(gen, observer, market, subnet, names, preload.op_mat_a, preload.op_mat_b,
                preload.one_step, options, date, forecast, solution['dispatch'])

    elif options['method'] =='planning':
        n_b = len(building)
        n_s = len(date)-1
        for t in range(n_s):
            dt_now = (date[t+1] - date[t]).seconds/3600
            if n_b > 0:
                setpoint = {}
                setpoint['temperature'] = [solution['building'][k]['temperature'][t] for k in names['buildings']]
                setpoint['district_heat'] = [solution['building'][k]['heating'][t]*1000 for k in names['buildings']]
                setpoint['district_cool'] = [solution['building'][k]['cooling'][t]*1000 for k in names['buildings']]
                weather_now = {}
                for i in actual_data['weather']:
                    weather_now[i] = actual_data['weather'][i][t]
                observer,net_electric = building_response(building,observer, weather_now, {},dt_now,setpoint,date[t])
                actual_data['building']['E0'][t] = net_electric
                actual_data['building']['C0'][t] = [solution['building'][k]['cooling'][t] for k in names['buildings']]
                actual_data['building']['H0'][t] = [solution['building'][k]['heating'][t] for k in names['buildings']]
            for i in range(len(observer['fluid_loop_temperature'])):
                gs_now = {}
                for k in solution['generator_state']:
                    gs_now[k] = solution['generator_state'][k][t]
                production = component_output(gen,subnet['cw'],gs_now)
                observer['fluid_loop_temperature'][i] += dt_now/observer['fluid_loop_capacitance'][i]*production[i]


def building_request(b_names,net,subnet,solution,forecast,t_sel):
    request = {}
    request['minimum'] = [[0] for n in range(len(b_names))]
    request['maximum'] = [[0] for n in range(len(b_names))]
    request['nominal'] = [[0] for n in range(len(b_names))]
    for i in range(len(b_names)):
        t_avg = solution[b_names[i]]['temperature'][0]
        if net == 'district_cool':
            t_bar = forecast['tc_bar'][0][i]
            ua = forecast['ua_c'][0][i]
            min_e = forecast['c_min'][0][i]
            t_min = forecast['Tmax'][0][i] #switched because Tmax should be temperature with minimum cooling
            t_max = forecast['Tmin'][0][i]
        elif net == 'district_heat':
            t_bar = forecast['th_bar'][0][i]
            ua = forecast['ua_h'][0][i]
            min_e = forecast['h_min'][0][i]
            t_min = forecast['Tmin'][0][i]
            t_max = forecast['Tmax'][0][i]
        request['nominal'][i] = max([(t_bar[0]-t_avg)*ua[0],(t_avg-t_bar[1])*ua[1],min_e])
        request['minimum'][i] = min([request['nominal'][i],max([(t_bar[0]-t_min)*ua[0],(t_min-t_bar[1])*ua[1],min_e])])
        request['maximum'][i] = max([request['nominal'][i],(t_bar[0]-t_max)*ua[0],(t_max-t_bar[1])*ua[1],min_e])
    return request


import numpy as np

def leward_calc(wind,surf_normal):
    threshold = 100 #degrees difference between wind and surface normal to be considered leward
    surf = [np.degrees(np.arctan2(surf_normal[i][0],surf_normal[i][1])) for i in range(len(surf_normal))]
    wind2 = wind-360
    leward = []
    for i in range(len(surf_normal)):
        if surf_normal[i][2]==1 or surf_normal[i][2]==-1:
            leward.append(True) 
        elif abs(wind-surf[i])>threshold and abs(wind2-surf[i])>threshold:
            leward.append(True)  
        else:
            leward.append(False)
    return leward


from datetime import timedelta

def build_time_vector(d1,options, to_timedelta=False):
    """Build time vector assuming constant timesteps.
    """
    n_steps = int(options['horizon'] / options['resolution'])
    time = [options['resolution']*t for t in range(n_steps+1)]
    if to_timedelta:
        time = [d1+timedelta(hours=x) for x in time]
    return time



'''
Linearly interpolates for one or more points regardless of whether x_vec is increasing or decreasing
'''

def two_way_interp(x_point,x_vec,y_vec):
    y_point = []
    if type(x_point)!=list:
        x_point = [float(x_point)]
    for xp in x_point:
        i = 0
        if x_vec[1]<x_vec[0]:
            while i<len(x_vec)-1 and xp<x_vec[i+1]:
                i +=1
        else:
            while i<len(x_vec)-1 and xp>x_vec[i+1]:
                i +=1
        if i == len(x_vec)-1:
            i-=1
        if i==0:
            y_point.append(y_vec[0])
        else:
            r = abs(x_vec[i+1]-xp)/abs(x_vec[i+1]-x_vec[i])
            y_point.append(r*y_vec[i] + (1-r)*y_vec[i+1])
    if len(y_point)==1:
        y_point = y_point[0]
    return y_point


def sort_solution(x, qp, names, gen, date, v_h):
    """
    Organizes the solution vector into a solution matrix where each column is
    for one component, and each row is for one timestep.

    INPUTS:
    qp          - instance of class QP of matrices
    x           - vector of unsorted optimal dispatch setpoints
    names       - dictionary of component, line, building and fluid_loop names
    gen         - list of generators and their attributes
    date        - list of datetime objects for the current and forecast times
    v_h         - value heat from combined heat and power generators when not excess heat

    OUTPUTS:
    solution    - sorted solution of optimal dispatch
    """

    n_s = len(date) - 1  # Numer of steps.
    n_g = len(qp['const_cost'])
    n_b = len(qp['organize']['building']['r'])
    n_l = len(qp['organize']['transmission'])
    
    solution = dict(
        timestamp = date,
        dispatch = [],
        generator_state = {},
        storage_state = {},
        excess_heat = {},
        excess_cool = {},
        line_flow = {},
        line_loss = {},
        building = {},
        fluid_loop = {},
    )
    solution['dispatch'] = sort_dispatch(x,qp)
    solution['generator_state'], solution['storage_state'] = gen_stor_state(gen,names['components'],solution['dispatch'],date)
    for i,k in enumerate(names['lines']):# Power transfer or water flow rate.
        solution['line_flow'][k] = [x[qp['indices'][n_g+i][t+1][0]] for t in range(n_s)]#flow leaving the upstream node (before losses accounted for)
        if len(qp['indices'][n_g+i][1]) > 1:
            solution['line_loss'][k] = [x[qp['indices'][n_g+i][t+1][1]] + x[qp['indices'][n_g+i][t+1][2]] for t in range(n_s)]#add loss terms in both directions
    for i,k in enumerate(names['buildings']):
        solution['building'][k] = {}
        solution['building'][k]['temperature'] = [x[qp['indices'][n_g+n_l+i][t+1][0]] for t in range(n_s)]
        solution['building'][k]['heating'] = [x[qp['indices'][n_g+n_l+i][t+1][1]] for t in range(n_s)]
        solution['building'][k]['cooling'] = [x[qp['indices'][n_g+n_l+i][t+1][2]] for t in range(n_s)]
        solution['building'][k]['heating'] = [v if abs(v)>1e-3 else 0 for v in solution['building'][k]['heating']]
        solution['building'][k]['cooling'] = [v if abs(v)>1e-3 else 0 for v in solution['building'][k]['cooling']]
    for i,k in enumerate(names['fluid_loop']):
        solution['fluid_loop'][k] = [x[qp['indices'][n_g+n_l+n_b+i][t+1][0]] for t in range(n_s)]

    sort_hydro(x,qp,solution,names['hydro'])
    for i,k in enumerate(names['heating_nodes']):
        solution['excess_heat'][k] = [x[qp['organize']['heat_vented'][i][t]] if qp['organize']['heat_vented'][i] else 0 for t in range(n_s)]
    for i,k in enumerate(names['cooling_nodes']):
        solution['excess_cool'][k] = [x[qp['organize']['cool_vented'][i][t]] if qp['organize']['cool_vented'][i] else 0 for t in range(n_s)]
    solution['value_heat'] = v_h
    return solution

def sort_hydro(x,qp,solution,names):
    n_s = qp['organize']['n_s']
    solution['hydro_soc'] = {}
    solution['hydro_spill'] =  {}
    solution['hydro_gen'] =  {}
    solution['hydro_out'] =  {}
    for i,k in enumerate(names):
    # Get SOC of each generator into a matrix for all time steps.
        solution['hydro_gen'][k] = [x[qp['organize']['ic'][qp['organize']['hydro'][i]]]]
        solution['hydro_soc'][k] = [x[qp['organize']['ic'][qp['organize']['hydro'][i]]+1]]
        solution['hydro_gen'][k].extend([x[qp['indices'][qp['organize']['hydro'][i]][t+1][0]] for t in range(n_s)])
        solution['hydro_soc'][k].extend([x[qp['indices'][qp['organize']['hydro'][i]][t+1][1]] for t in range(n_s)])
        solution['hydro_out'][k] = [x[qp['indices'][qp['organize']['hydro_dr_line']][t+1]] for t in range(n_s)]
        if 'hydro_soc_offset' in qp['organize']:
            solution['hydro_soc'][k] = [j + qp['organize']['hydro_soc_offset'][i] for j in solution['hydro_soc'][k]]
        if len(qp['indices'][qp['organize']['hydro'][i]][1]) ==3 or len(qp['indices'][qp['organize']['hydro'][i]][1]) == 5: #has spill flow
            solution['hydro_spill'][k] = [x[qp['indices'][qp['organize']['hydro'][i]][t+1][2]] for t in range(n_s)]
        else:
            solution['hydro_spill'][k] = [0 for t in range(n_s)]
        

def sort_dispatch(x,qp):
    n_g = len(qp['const_cost'])
    n_b = len(qp['organize']['building']['r'])
    n_l = len(qp['organize']['transmission'])
    n_fl = 0
    if 'cooling_water' in qp['organize']['balance']:
        n_fl = len(qp['organize']['balance']['cooling_water'])
    dispatch = [[] for i in range(n_g + n_l + n_b + n_fl)]
    if 'n_s' in qp['organize']:
        n_s = qp['organize']['n_s']
        for i in range(n_g):
            dispatch[i] = sort_gen_disp(x, qp, i)
        for i in range(n_l):# Power transfer or water flow rate.
            dispatch[n_g+i].append(0)
            dispatch[n_g+i].extend([x[qp['indices'][n_g+i][t+1][0]] for t in range(n_s)])
        for i in range(n_b):
            dispatch[n_g+n_l+i].append(0)
            dispatch[n_g+n_l+i].extend([x[qp['indices'][n_g+n_l+i][t+1][0]] for t in range(n_s)])
        for i in range(n_fl):
            dispatch[n_g+n_l+n_b+i].append(0)
            dispatch[n_g+n_l+n_b+i].extend([x[qp['indices'][n_g+n_l+n_b+i][t+1][0]] for t in range(n_s)])
    else:
        dispatch[:n_g] = sort_gen_disp_step(x, qp)
        dispatch[n_g:n_g+n_l] = [x[qp['indices'][n_g+i][0][0]] for i in range(n_l)]# Power transfer or water flow rate.
        dispatch[n_g+n_l:n_g+n_l+n_b] = [x[qp['indices'][n_g+n_l+i][0][0]] for i in range(n_b)]
        dispatch[n_g+n_l+n_b:n_g+n_l+n_b+n_fl] = [x[qp['indices'][n_g+n_l+n_b+i][0][0]] for i in range(n_fl)]
    return dispatch

def sort_gen_disp(x, qp, i):
    n_s = qp['organize']['n_s']
    dispatch = []
    if not qp['indices'][i][-1] is None and not qp['indices'][i][-1][0] is None:
        # Linear mapping between state and output.
        out_vs_state = qp['organize']['out_vs_state'][i]
        if not qp['indices'][i][0][0] is None and x[qp['indices'][i][0][0]] > 2e-4:#Initial conditions
            dispatch.append(x[qp['indices'][i][0][0]])
        else:
            dispatch.append(0)
        for t in range(n_s):
            p = sum([out_vs_state[k] * x[s] for k,s in enumerate(qp['indices'][i][t+1])])# Record this combination of outputs (SOC for storage).
            if abs(p) < 2e-4:
                p = 0  # Added to avoid rounding errors in optimization.
            dispatch.append(p)
    elif isinstance(qp['renewable'][i],(float,int)) or len(qp['renewable'][i])>0:
        dispatch.append(0)
        dispatch.extend(qp['renewable'][i])
    else:
        dispatch = [0 for t in range(n_s+1)]
    return dispatch

def sort_gen_disp_step(x, qp):
    n_g = len(qp['const_cost'])
    dispatch = []
    for i in range(n_g):
        if not qp['indices'][i][-1] is None and not qp['indices'][i][-1][0] is None:
            # Linear mapping between state and output.
            out_vs_state = qp['organize']['out_vs_state'][i]
            p = sum([out_vs_state[k] * x[s] for k,s in enumerate(qp['indices'][i][0])])# Record this combination of outputs (SOC for storage).
            if abs(p) < 2e-4:
                p = 0  # Added to avoid rounding errors in optimization.
            dispatch.append(p)
        elif isinstance(qp['renewable'][i],(float,int)) or len(qp['renewable'][i])>0:
            dispatch.append(qp['renewable'][i])
        else:
            dispatch.append(0)
    return dispatch

def sort_eh(x, qp):
    if 'n_s' in qp['organize']:
        n_s = qp['organize']['n_s']
    n_dh = len(qp['organize']['heat_vented'])
    excess_heat = [[] for n in range(n_dh)]
    for n in range(n_dh):
        if 'n_s' in qp['organize']:
            if qp['organize']['heat_vented'][n]:
                excess_heat[n] = [x[qp['organize']['heat_vented'][n][t]] for t in range(n_s)]
            else:
                excess_heat[n] = [0 for t in range(n_s)]
        else:
            if qp['organize']['heat_vented'][n]:
                excess_heat[n] = x[qp['organize']['heat_vented'][n][0]]
            else:
                excess_heat[n] = 0
    return excess_heat

def sort_ec(x,qp):
    if 'n_s' in qp['organize']:
        n_s = qp['organize']['n_s']
    n_dc = len(qp['organize']['cool_vented'])
    excess_cool = [[] for n in range(n_dc)]
    for n in range(n_dc):
        if 'n_s' in qp['organize']:
            if qp['organize']['cool_vented'][n]:
                excess_cool[n] = [x[qp['organize']['cool_vented'][n][t]] for t in range(n_s)]
            else:
                excess_cool[n] = [0 for t in range(n_s)]
        else:
            if qp['organize']['cool_vented'][n]:
                excess_cool[n] = x[qp['organize']['cool_vented'][n][0]]
            else:
                excess_cool[n] = 0
    return excess_cool


def sort_building_step(x,qp):
    n_g = len(qp['const_cost'])
    n_b = len(qp['organize']['building']['r'])
    n_l = len(qp['organize']['transmission'])
    s_build = dict(
            heating = [[] for i in range(n_b)],
            cooling = [[] for i in range(n_b)],
            temperature = [[] for i in range(n_b)],)
    for i in range(n_b):
        bs = qp['indices'][n_g+n_l+i][0]
        s_build['temperature'][i] = x[bs[0]]
        s_build['heating'][i] = x[bs[1]]
        s_build['cooling'][i] = x[bs[2]]
    return s_build

def gen_stor_state(gen,names,dispatch,date):
    n_s = len(date) -1
    dt = [(date[t+1] - date[t]).total_seconds()/3600 for t in range(n_s)]# Duration of each time segment in hours.
    gs = {}
    ss = {}
    for i,k in enumerate(names):
        if 'stor' in gen[i]:
            ss[k] = [dispatch[i][t+1] for t in range(n_s)]
            sd = soc_2_power(gen[i]['stor'],dispatch[i],dt)
            gs[k] = sd
        else:
            gs[k] = [dispatch[i][t+1] for t in range(n_s)]
    return gs,ss

def soc_2_power(gen,disp,dt):
    ss = [i for i in disp]
    sd = []
    for t in range(len(dt)):
        loss = gen['self_discharge'] * gen['usable_size']*dt[t] # %/hr * capacity (kWh) * time in hours = kWh lost
        d_soc = (ss[t+1] - ss[t]) + loss #kWh add to storage (positive is charging)
        if ss[t+1]>ss[t]:
            sd.append(-d_soc/gen['charge_eff']/dt[t]) # power in kW
        else:
            sd.append(-d_soc*gen['disch_eff']/dt[t]) # power in kW
    return sd


from eagers.basic.hdf5 import DatetimeFloatConverter as DFC

def weather_forecast(data, prev_data, hist_prof, date):
    """Weather forecast."""
    forecast = {}
    do_not_overwrite = []
    if 'forecasts' in data and 'weather' in data.forecasts:
        f = list(data.forecasts.weather.keys())
        for k in f:
            forecast[k] = data.forecasts.weather[k]
            do_not_overwrite.append(k)
    if 'weather' in data:
        p_time = DFC.d2f_arr2arr(prev_data['timestamp'])
        hour = date.hour()
        nd = len(hour)
        n_time = DFC.d2f_arr2arr(date)
        for f in prev_data:
            if f == 'timestamp':
                forecast[f] = date
            if f in do_not_overwrite or not isinstance(prev_data[f][0],(float,int)):
                pass
            else:
                if len(date) == 1:
                    forecast[f] = prev_data[f][-1] #most recent data point
                else:
                    ip = 0
                    for t in range(nd):
                        while ip<len(p_time) and p_time[ip]<n_time(t):
                            ip+=1
                        if ip == 0:
                            yf = prev_data[f][ip]
                        else: 
                            r = (n_time[t]-p_time[ip])/(p_time[ip]-p_time[ip-1])
                            yf = r*prev_data[f][ip-1]+(1-r)*prev_data[f][ip]
                        if hour[t]<1:
                            hf = (1-hour[t])*hist_prof[f][-1]+hour[t]*hist_prof[f][0]
                        else:
                            hp = round(hour[t]-.5,0) # round down to nearest hour
                            r = hour[t] - hp
                            hf = (1-r)*hist_prof[f][hp]+hour[t]*hist_prof[f][hp+1]
                        offset = prev_data[f][-1] - prev_data[f][0] 
                        if hour[t] <1:
                            forecast[f][t] = (1-hour[t]/10)*(prev_data[f][-1])+hour[t]/10*hf
                        if hour[t]<4:
                            forecast[f][t] = (1.1-hour[t]/5)*(yf+offset)+(hour[t]/5-.1)*hf #Balanced between yesterdays T and historical T (includes forecast for current time)
    return forecast



"""Fix infeasible result.
"""

from copy import copy, deepcopy

from eagers.solver.ecos_qp import ecos_qp


def fix_infeasible(qp):
    qp_new = {}
    # Copy 1-D lists and deepcopy nested lists.
    qp_new['h'] = copy(qp['h'])
    qp_new['f'] = copy(qp['f'])
    qp_new['a'] = deepcopy(qp['a'])
    qp_new['b'] = copy(qp['b'])
    qp_new['a_eq'] = deepcopy(qp['a_eq'])
    qp_new['b_eq'] = copy(qp['b_eq'])
    qp_new['ub'] = copy(qp['ub'])
    qp_new['lb'] = copy(qp['lb'])
    qp_new['x_keep'] = copy(qp['x_keep'])
    qp_new['r_keep'] = copy(qp['r_keep'])
    qp_new['req_keep'] = copy(qp['req_keep'])
    n_s = qp['organize']['n_s']
    node_bal = qp['organize']['balance']
    bal = dict(num = [], row = [], net = [], time = [], node = [])
    b = 0
    f_name = list(qp['organize']['balance'].keys())
    for net in f_name:
        for j,row in enumerate(node_bal[net]):
            bal['num'].extend([b+t for t in range(n_s)])
            bal['row'].extend(row) 
            bal['net'].extend([net] * n_s)
            bal['time'].extend([t+1 for t in range(n_s)])
            node_name = qp['network'][net]['node_names'][j]
            bal['node'].extend([node_name for t in range(n_s)])
            b= bal['num'][-1]+1
    s_add = sum([sum([len(n) for n in node_bal[net]]) for net in f_name])
    z_add = [0 for i in range(s_add)]
    r = len(qp_new['a'])
    r_eq = len(qp_new['a_eq'])
    x_l = len(qp_new['f'])
    qp_new['h'].extend(z_add)
    max_c = 1e3 * max(qp_new['f'])
    qp_new['f'].extend([max_c for i in range(s_add)])
    for rq in range(r_eq):
        qp_new['a_eq'][rq].extend(z_add)
    for rnq in range(r):
        qp_new['a'][rnq].extend(z_add)
    qp_new['lb'].extend(z_add)
    qp_new['ub'].extend([1e10 for i in range(s_add)])
    qp_new['x_keep'].extend([True for i in range(s_add)])
    for j in range(len(bal['row'])):
        qp_new['a_eq'][bal['row'][j]][x_l+bal['num'][j]] = 1

    x, feasible = ecos_qp(qp_new['h'], qp_new['f'], qp_new['a'], qp_new['b'], qp_new['a_eq'], qp_new['b_eq'], qp_new['ub'], qp_new['lb'], qp_new['x_keep'], qp_new['r_keep'], qp_new['req_keep'])
    x = x[:len(qp['f'])]
    if feasible != 0:
        print('Some additional contraints making problem infeasible.\n' \
            + 'Tried adding infinite sources.')
    else:
        node_makeup = [[val,i] for i,val in enumerate(x[x_l:]) if abs(val)>1e-3]
        n_mku = len(node_makeup)
        if n_mku > 0:
            if n_mku <= 4:
                # Display:
                # - The energy type for which production was
                #   insufficient,
                # - The node at which this occurred,
                # - The time interval into the future at which this
                #   occurred.
                for i in range(n_mku):
                    j = bal['num'].index(node_makeup[i][1])
                    net = bal['net'][j]
                    v = node_makeup[i][0]
                    t = bal['time'][j]
                    node = bal['node'][j]
                    print(
                        f'Short production capacity of {net} '
                        f'By approximately {v}'
                        f'at node {node} '
                        f'at time interval {t}.'
                    )
            else:
                short_net = dict(net = [],ti = [])
                for i in range(n_mku):
                    j = bal['num'].index(node_makeup[i][1])
                    if not bal['net'][j] in short_net:
                        short_net['net'].append(bal['net'][j])
                        short_net['ti'].append(1)
                    else:
                        n = short_net['net'].index(bal['net'][j])
                        short_net['ti'][n] +=1
                for f in range(len(short_net['net'])):
                    net = short_net['net'][f]
                    n = short_net['ti'][f]
                print(f'Short production capacity of {net}')
    return x, feasible



from eagers.building.components.manager_override import manager_override
from eagers.building.basic.eval_curve import eval_curve
from eagers.building.components.water_heater import water_heater
from eagers.building.components.cooling_tower import cooling_tower
from eagers.building.process.flow_resolver_supply import boiler_PLR
from eagers.building.process.flow_resolver_supply import chiller_PLR

def supply_loop(building,profile,plant_nodes,wh_tank,T_air,w_air,l,dt):
    Cp_water = 4186  #J/kg*K
    fresh_air = {}
    fresh_air['Tdb'] = T_air 
    fresh_air['w'] = w_air 
    cool_elec = 0
    heat_gas = 0
    water_gas = 0
    water_elec = 0
    tower_elec = 0
    s_equip = building.plant_supply_equip 
    d_equip = building.plant_demand_equip 
    on_loop = [i for i,x in enumerate(s_equip['loop']) if x == l]
    for j in on_loop:
        T_target = building.plant_loop['exit_temperature'][l] 
        node = building.plant_loop['setpoint_node'][l] 
        T_target = manager_override(building,profile,node,T_target,None,fresh_air,None) 
        in_n = s_equip['inlet'][j] 
        out_n = s_equip['outlet'][j] 
        if s_equip['type'][j]=='pipe:adiabatic':
            plant_nodes['supply_temperature'][out_n] = plant_nodes['supply_temperature'][in_n] 
        elif s_equip['type'][j]=='mixer':
            if type(in_n) !=list:
                in_n = [in_n]
            if plant_nodes['supply_flow'][out_n]>0:
                plant_nodes['supply_temperature'][out_n] = sum([plant_nodes['supply_temperature'][k]*plant_nodes['supply_flow'][k]/plant_nodes['supply_flow'][out_n] for k in in_n])
            else: #avoid divide by zero
                plant_nodes['supply_temperature'][out_n] = sum([plant_nodes['supply_temperature'][k] for k in in_n]) 
        elif s_equip['type'][j]=='splitter':
            if type(out_n) !=list:
                out_n = [out_n]
            for k in out_n:
                plant_nodes['supply_temperature'][k] = plant_nodes['supply_temperature'][in_n] 
        elif s_equip['type'][j]=='chiller:electric:reformulatedeir' or s_equip['type'][j]=='chiller:electric:eir':
            PLR,Q_evap,plant_nodes['supply_temperature'][out_n],Q_avail,c = chiller_PLR(building,s_equip['name'][j],plant_nodes['supply_flow'][in_n],plant_nodes['supply_temperature'][in_n],plant_nodes['supply_temperature'][out_n])
            if PLR > 0:
                cycling_ratio = min(PLR/building.chiller[c]['min_part_load'],1) 
                PLR = max(PLR,building.chiller[c]['min_unload_ratio']) 
                Q_false_load = Q_avail*PLR*cycling_ratio - Q_evap 
                if 'electric_in_out_curve_type' in building.chiller[c]:
                    if building.chiller[c]['electric_in_out_curve_type']=='leavingcondenserwatertemperature':
                        C_EIR_PLR = eval_curve(building.curves,building.chiller[c]['electric_in_out_part_load_curve'],[building.chiller[c]['condenser_water_temperature_ref'],PLR]) 
                    else:
                        dT_ref = building.chiller[c]['condenser_water_temperature_ref'] - building.chiller[c]['chilled_water_temperature_ref']
                        dT = (building.chiller[c]['condenser_water_temperature_ref'] - plant_nodes['supply_temperature'][out_n])/dT_ref 
                        T_dev = (plant_nodes['supply_temperature'][out_n] - building.chiller[c]['chilled_water_temperature_ref'])/dT_ref 
                        C_EIR_PLR = eval_curve(building.curves,building.chiller[c]['electric_in_out_part_load_curve'],[dT,PLR,T_dev]) 
                else:
                    C_EIR_PLR = eval_curve(building.curves,building.chiller[c]['electric_in_out_part_load_curve'],[PLR]) 
                CEIR_T = eval_curve(building.curves,building.chiller[c]['electric_in_out_temperature_curve'],[plant_nodes['supply_temperature'][out_n],building.chiller[c]['condenser_water_temperature_ref']]) 
                P_chiller = PLR*Q_avail*(1/building.chiller[c]['COP_ref'])*CEIR_T*C_EIR_PLR*cycling_ratio 
                if building.chiller[c]['frac_power_2_condenser']== None:
                    Q_cond = Q_evap+Q_false_load
                else:
                    Q_cond = (P_chiller*building.chiller[c]['frac_power_2_condenser'])+Q_evap+Q_false_load 
                if 'condenser_type' in building.chiller[c] and building.chiller[c]['condenser_type']=='aircooled':
                    pass
                    ## Do something for air_cooled chiller?
                else:#water cooled
                    d_e_num = d_equip['name'].index(s_equip['name'][j])
                    plant_nodes['load'][d_equip['outlet'][d_e_num]] = Q_cond 
                    plant_nodes['demand_flow'][d_equip['outlet'][d_e_num]] = Q_cond/(Cp_water*building.plant_loop['temperature_difference'][l]) 
                cool_elec += P_chiller #Watts
        elif s_equip['type'][j]=='boiler:hotwater':
            PLR,load,plant_nodes['supply_temperature'][out_n],b = boiler_PLR(building,s_equip['name'][j],plant_nodes['supply_flow'][in_n],plant_nodes['supply_temperature'][in_n],plant_nodes['supply_temperature'][out_n])
            if load>0:
                if building.boiler[b]['normalized_eff_curve']==None:
                    eff = building.boiler[b]['efficiency']
                else:
                    try:
                        eff = eval_curve(building.curves,building.boiler[b]['normalized_eff_curve'],[PLR,plant_nodes['supply_temperature'][out_n]]) 
                    except ValueError:
                        print('error in manage_plant_equip, supply_loop')
                heat_gas += load/eff #Watts
        elif s_equip['type'][j]=='pump:variablespeed' or s_equip['type'][j]=='pump:constantspeed':
            ## pump heat goes into tank at start of half-loop
            plant_nodes['supply_temperature'][out_n] = plant_nodes['supply_temperature'][in_n] 
        elif s_equip['type'][j]=='waterheater:mixed':
            T_exit = profile[building.water_heater[s_equip['name'][j]]['temperature_schedule']]
            wh_tank[s_equip['name'][j]]['T'],wh_tank[s_equip['name'][j]]['on'],wh_gas,wh_elec = water_heater(building,wh_tank[s_equip['name'][j]]['T'],wh_tank[s_equip['name'][j]]['on'],T_exit,T_air,plant_nodes['supply_temperature'][in_n],plant_nodes['supply_flow'][in_n],dt,s_equip['name'][j]) 
            water_gas += wh_gas
            water_elec += wh_elec
            plant_nodes['supply_temperature'][out_n] = wh_tank[s_equip['name'][j]]['T'] 
        elif (s_equip['type'][j]=='coolingtower:singlespeed' or
              s_equip['type'][j]=='coolingtower:twospeed' or
              s_equip['type'][j]=='coolingtower:variablespeed:merkel'):
            P_fan,plant_nodes['supply_temperature'][out_n],bypass = cooling_tower(building,s_equip['name'][j],plant_nodes['supply_flow'][in_n],plant_nodes['supply_temperature'][in_n],T_target,T_air,w_air) 
            tower_elec += P_fan #Watts
        else:
            print('need additional plant supply loop component__' + s_equip['type'][j])
    return plant_nodes,wh_tank,cool_elec,heat_gas,water_gas,water_elec,tower_elec


"""Write results from HDF5 to CSV format.

Functions:
write_result_to_csv
"""

import csv

import tables as tb

from eagers.config.path_spec import CSV_SUFFIX, USER_DIR_SIMRESULTS
from eagers.config.text import CSV_RESULT_SEP
from eagers.write.result_file import existing_h5_result_filepath
from eagers.basic.file_handling import ensure_suffix
from eagers.basic.hdf5 import h5file_context


def write_result_to_csv(proj_name):
    """Write result from HDF5 to CSV format."""
    h5path = existing_h5_result_filepath(proj_name)
    # Open file in 'r'ead mode, requiring its existence already.
    with h5file_context(h5path, 'r') as h5f:
        # Read data from result tables.
        root = h5f.root
        tables = list(c for c in root._v_children
            if isinstance(getattr(root, c), tb.table.Table))
        data = {table: getattr(root, table).read() for table in tables}
        # Define function for walking the fields of a structured array
        # dtype.
        def walk_dtype_fields(dtype, path, base_yield):
            fields = dtype.fields
            # "is not None" accounts for dtypes with 0 fields.
            if fields is not None:
                for k, v in fields.items():
                    subp = path + f"/{k}"
                    yield from walk_dtype_fields(v[0], subp, base_yield)
            else:
                yield base_yield(path, dtype)
        # Define function for getting data at a given path.
        def data_at_path(data, path):
            split_path = path.split('/')[1:]
            d = data
            for p in split_path:
                d = d[p]
            return d
        # Write to CSV file.
        def write_to_csv(proj_name, table_name):
            filename = f"{proj_name}{CSV_RESULT_SEP}{table}"
            csvpath = USER_DIR_SIMRESULTS / ensure_suffix(filename, CSV_SUFFIX)
            headers = list(walk_dtype_fields(
                data[table].dtype, "", (lambda p, d: p)))
            with open(csvpath, 'w', newline='') as csvf:
                writer = csv.DictWriter(csvf, fieldnames=headers)
                writer.writeheader()
                for entry in data[table]:
                    writer.writerow(dict(walk_dtype_fields(
                        data[table].dtype,
                        "",
                        (lambda p, d: (p, data_at_path(entry, p))),
                    )))
        for table in tables:
            write_to_csv(proj_name, table)



'''
Converts the air flow and temperature to heat/cooling loads on water loops 
and to electric/gas loads if it is directly to an electric/gas component
'''

from eagers.building.basic.psychometric import psychometric
from eagers.building.components.water_coil import water_coil
from eagers.building.components.cooling_coil import cooling_coil
from eagers.building.components.fan_calc import fan_calc
from eagers.building.components.unitary_heat_cool import unitary_heat_cool

def sim_air_loops(building,air_nodes,mixed_air_loop,central_flow,w_loop,T_air,m_v_air,plant_nodes,e_use,fan_avail):
    supply_air = {}
    for k in mixed_air_loop:
        supply_air[k] = [j for j in mixed_air_loop[k]]
    for l in range(len(supply_air['Tdb'])):  #loop through equipment on loop
        on_loop = [i for i,x in enumerate(building.air_supply_equip['loop']) if x==l]
        air = {}
        air['Tdb'] = supply_air['Tdb'][l]+0
        air['w'] = supply_air['w'][l]+0
        air['m_dot'] = supply_air['m_dot'][l]+0
        air['h'] = supply_air['h'][l]+0 #update intermediate air as it passes between equipment
        for j in range(len(on_loop)):
            ## making sure simulation of these components is in the order they are on the supply branch
            k = on_loop[[building.air_supply_equip['branch_order'][on_loop[i]] for i in range(len(on_loop))].index(j)]
            name = building.air_supply_equip['name'][k]
            out = building.air_supply_equip['outlet'][k]
            if building.air_supply_equip['type'][k]=='airloophvac:outdoorairsystem':
                # do nothing, already mixed air
                pass
            elif (building.air_supply_equip['type'][k]== 'coil:cooling:water:detailedgeometry' or
                 building.air_supply_equip['type'][k]== 'coil:cooling:water' or
                 building.air_supply_equip['type'][k]== 'coil:heating:water'):
                air,m_w,actual_load,out = water_coil(building,name,air,air_nodes['supply_T_set'][out],plant_nodes)
                plant_nodes['demand_flow'][out] = m_w+0 #kg/s
                plant_nodes['load'][out] = actual_load+0 #load in Watts (negative for heating coil because air leaves hotter)
            elif (building.air_supply_equip['type'][k]=='coilsystem:cooling:dx' or
                 building.air_supply_equip['type'][k]=='dx:twospeed' or
                 building.air_supply_equip['type'][k]=='dx:singlespeed'):
                T_wb = psychometric({'Tdp':T_air,'w':m_v_air},'Twb')
                air,P_coil = cooling_coil(building,name,air,T_air,air_nodes['supply_T_set'][out])
                e_use['cool_elec'] += P_coil
            elif (building.air_supply_equip['type'][k]=='coil:heating:fuel' or
                 building.air_supply_equip['type'][k]== 'coil:heating:electric'):
                spec_heat_supply = 1006 + 1860*air['w'] #J/kg*K
                Q_treat = air['m_dot']*spec_heat_supply*(air_nodes['supply_T_set'][out]-air['Tdb'])
                if Q_treat>1:
                    air['Tdb'] = air_nodes['supply_T_set'][out]+0
                    air['h'] = air['h'] + Q_treat/air['m_dot']
                    if building.air_supply_equip['type'][k]=='coil:heating:fuel':
                        e_use['heat_gas']+= Q_treat/building.coils_heating[name]['efficiency'] #heat added in W
                    else:
                        e_use['heat_elec']+= Q_treat/building.coils_heating[name]['efficiency'] #heat added in W
            elif building.air_supply_equip['type'][k][:4] == 'fan:':
                air,P_fan,_ = fan_calc(building.fans[name],air,fan_avail[name])
                if ('end_use' in building.fans[name] and building.fans[name]['end_use'] == 'fan energy'):
                    e_use['fan_elec'] += P_fan
                else: # 'Unit Heater Fans', 'Zone Exhaust Fans'c
                    e_use['other_fans'] += P_fan
            elif building.air_supply_equip['type'][k]=='humidifier:steam:electric':
                if w_loop[l]>air['w']: #see page 1250
                    #steam specific enthalpy = 2676125 J/kg
                    den_water = 995 #kg/m^3
                    m_w = (w_loop[l]-air['w'])*air['m_dot']
                    air['h'] = air['h'] + m_w/air['m_dot']*2676125
                    air['Tdb'] = psychometric(air,'Tdp')
                    ## need to check it is not over saturated
                    k =building.humidifiers['name'].index(name)
                    e_use['heat_elec'] += m_w/(den_water*building.humidifiers['flow_capacity'][k])*building.humidifiers['max_power'][k] + building.humidifiers['fan_power'][k] + building.humidifiers['standby_power'][k]
            elif building.air_supply_equip['type'][k]=='airloophvac:unitaryheatcool':
                air,plant_nodes,e_use = unitary_heat_cool(building,fan_avail,air,air_nodes['supply_T_set'][out],T_air,e_use,plant_nodes,building.unitary_heat_cool[name])
            else:
                print('need to add components in sim_air_loops')
        supply_air['Tdb'][l] = air['Tdb']+0
        supply_air['w'][l] = air['w']+0
        supply_air['m_dot'][l] =air['m_dot']+0
        supply_air['h'][l] = air['h']+0 #update intermediate air as it passes between equipment

    air_density = 1.204 #kg/m^3 %standard air density (1.204 kg/m3) adjusted for the local barometric pressure (standard barometric pressure corrected for altitude, ASHRAE 1997 HOF pg. 6.1).
    n_z = len(building.zones['name'])
    zone_air = {}
    zone_air['Tdb'] = [0 for i in range(n_z)]
    zone_air['w'] = [0 for i in range(n_z)]
    zone_air['h'] = [0 for i in range(n_z)]
    zone_air['flow'] = [j for j in central_flow]
    zone_air['m_dot'] = [air_density*central_flow[i] for i in range(n_z)]

    for z in range(n_z):
        l = building.zones['air_loop'][z]
        if l !=None:
            zone_air['Tdb'][z] = supply_air['Tdb'][l]+0
            zone_air['w'][z] = supply_air['w'][l]+0
            zone_air['h'][z] = supply_air['h'][l]+0

    ## exhaust fan power
    for i in building.fans:
        if building.fans[i]['exhaust'] and building.fans[i]['flow_rate']!=None:
            fan_flow = building.fans[i]['flow_rate']*fan_avail[i] #volumetric flow (m^3/s)  %See page 1601 of the reference manual
            P_fan += fan_flow*building.fans[i]['pressure_rise']/(building.fans[i]['fan_efficiency']) # m3/s * Pa = N*m/s = W
            if ('end_use' in building.fans[i] and building.fans[i]['end_use'] == 'fan energy'):
                e_use['fan_elec'] += P_fan
            else: # 'Unit Heater Fans', 'Zone Exhaust Fans'c
                e_use['other_fans'] += P_fan
    return zone_air,e_use,plant_nodes


import datetime as dt

from eagers.basic.hdf5 import DatetimeFloatConverter as DFC

def typical_day(timestamp, data):
    timestamp = DFC.f2d_arr2lst(timestamp)#convert float back to datetime
    x = [[[] for j in range(24)] for i in range(12)]
    z = [[[] for j in range(24)] for i in range(12)]
    for i, t in enumerate(timestamp):
        x[t.month][t.hour].append(data[i])
    for m in range(12):
        if all([len(x[m][h])>0 for h in range(24)]): #data exists for every hour
            for h in range(24):
                if any(x[m][h]):
                    z[m][h] = sum(x[m][h])/len(x[m][h]) #average at this hour
                else:
                    z[m][h] = 0
    not_empty = [m2 for m2 in range(12) if len([h2 for h2 in range(24) if len(z[m2][h2])>0])>0]
    for m in range(12):
        if not m in not_empty:
            a = [min([abs(m-m2),abs(m-12 - m2),abs(m+12-m2)]) for m2 in not_empty]
            minpos = a.index(min(a)) #find nearest non-empty month
            z[m] = z[not_empty[minpos]]    
        not_empty2 = [h2 for h2 in range(24) if len(z[m][h2])>0]
        for h in range(24):
            if not h in not_empty2:
                a = [min([abs(h-h2),abs(h-24 - h2),abs(h+24-h2)]) for h2 in not_empty2]
                minpos = a.index(min(a)) #find nearest non-empty hour
                z[m][h] = z[m][not_empty2[minpos]]
    return z



from eagers.basic.component_output import component_output
from eagers.simulate.dispatch_network import dispatch_network

def assign_disparity(net,names,gen,subnet,production,demand,request,solution,marginal,dt,tol,excess,re_opt):
    #check if there is a disparity with forecast
    #check if there is a utility (slack bus)
    #if not check if storage has enough capacity to be slack bus
    #if it is heating and cooling, assign disparity to buildings hot/cold water loops
    #otherwise re-optimize
    i_utility = []
    i_storage = []
    capacity = []
    thermal = None
    acdc = any([gen[i]['type']=='ACDCConverter' for i in range(len(gen))])
    if acdc:
        acdc_i = [i for i in range(len(gen)) if gen[i]['type']=='ACDCConverter']
        ac2dc = gen[acdc_i[0]]['output']['dc'][-1][0]
        dc2ac = gen[acdc_i[0]]['output']['e'][-1][1]
    for i in range(len(gen)):
        if subnet['abbreviation'] in gen[i]['output']:
            if gen[i]['type'] == 'Utility':
                    i_utility.append(i)
            elif gen[i]['type'] in ['ThermalStorage', 'ElectricStorage','HydroStorage','HydrogenStorage']:
                    i_storage.append(i)
        if gen[i]['type'] == 'ElectricStorage' and acdc and ((net == 'electrical' and 'dc' in gen[i]['output']) or (net == 'direct_current' and 'e' in gen[i]['output'])):
            i_storage.append(i)
        if gen[i]['type'] == 'Utility' and acdc and ((net == 'electrical' and 'dc' in gen[i]['output']) or (net == 'direct_current' and 'e' in gen[i]['output'])):
            i_utility.append(i)
    ##TODO solve a nodal problem to determine any shortfall/excess energy in the network
    prod = sum(production)
    req = sum([demand[k] for k in demand])
    if not request is None and 'nominal' in request:
        req += sum(request['nominal'])
        thermal = [j*1000 for j in request['nominal']]
    e_error = None
    if ((prod == 0 and req>0) or (prod>0 and abs(prod-req)/prod>tol)):
        e_error = req - prod #Positive means there is unmet demand, negative means excess production
        if e_error<0 and excess: #can dump heat or cooling
            e_error = None
    ######
    # e_error = dispatch_network(names,subnet,production,demand,request,marginal,capacity,True)
    ######
    
    if not e_error is None:
        if len(i_utility)==0 and len(i_storage)==0:
            re_opt = True
        elif len(i_storage)==0:  #no storage, grid handles error
            i = i_utility[0]
            if subnet['abbreviation'] in gen[i]['output']:
                solution['dispatch'][i][1] += e_error
            elif acdc and net == 'electrical':
                solution['dispatch'][i][1] += e_error/dc2ac
                solution['dispatch'][acdc_i[0]][1] -= e_error/dc2ac
            elif acdc and net == 'direct_current':
                solution['dispatch'][i][1] += e_error/ac2dc
                solution['dispatch'][acdc_i[0]][1] += e_error/ac2dc
            solution['generator_state'][gen[i]['name']][0] = solution['dispatch'][i][1]
        else: # add up storage capacity to meet error & split proportional to available capacity            
            cap = [0 for i in range(len(i_storage))]
            for i in range(len(i_storage)):
                k = i_storage[i]
                gs = gen[k]['stor']
                if e_error>0: #extra demand, what stored charge can be called upon
                    cap[i] = min(gs['peak_disch'],gs['disch_eff']*(gs['usable_size'] - solution['dispatch'][k][1])/dt)
                else: #less demand, what remaining storage capacity can be called upon
                    cap[i] = -min(gs['peak_charge'],(1/gs['charge_eff'])*solution['dispatch'][k][1]/dt)
                if subnet['abbreviation'] not in gen[i]['output'] and acdc and net == 'electrical':
                    cap[i] = cap[i]*dc2ac
                elif subnet['abbreviation'] not in gen[i]['output'] and acdc and net == 'direct_current':
                    cap[i] = cap[i]*ac2dc
            if len(i_utility)==0: #no grid, all error is absorbed by storage (if possible? otherwise re-optimize)
                if abs(sum(cap))<abs(e_error): #if unable to meet error, re-optimize
                    re_opt = True
                else:
                    r = abs(e_error)/abs(sum(cap))
                    if e_error>0: #extra demand, what stored charge can be called upon
                        power = r*cap[i]/gs['disch_eff']
                    else: #less demand, what remaining storage capacity can be called upon
                        power = r*cap[i]*gs['charge_eff']
                    if subnet['abbreviation'] not in gen[i]['output'] and acdc and net == 'electrical':
                        power = power/dc2ac
                        solution['dispatch'][acdc_i[0]][1] -= power
                        solution['generator_state'][gen[acdc_i[0]]['name']][0] -= power
                    elif subnet['abbreviation'] not in gen[i]['output'] and acdc and net == 'direct_current':
                        power = power/ac2dc
                        solution['dispatch'][acdc_i[0]][1] += power
                        solution['generator_state'][gen[acdc_i[0]]['name']][0] += power
                    solution['dispatch'][k][1] -= power*dt
                    solution['generator_state'][gen[k]['name']][0] += power
                    solution['storage_state'][gen[k]['name']][0] -= power*dt
            else: #split the error between utility and energy storage based on storage behavior
                charging = []
                discharging = []
                for k in i_storage:
                    ##TODO factor in loss term
                    if solution['dispatch'][k][1]>solution['dispatch'][k][0]: #charging
                        charging.append((1/gen[k]['stor']['charge_eff'])*(solution['dispatch'][k][1] - solution['dispatch'][k][0])/dt)
                        discharging.append(0)
                    else:
                        charging.append(0)
                        discharging.append(gen[k]['stor']['disch_eff']*(solution['dispatch'][k][0] - solution['dispatch'][k][1])/dt)
                r = 0
                if e_error>0 and sum(charging)>0: #If more demand than anticipated, reduce all storage charging, otherwise resort to grid
                    r = 1 - max([0,(sum(charging)-e_error)])/sum(charging)
                    e_error -= r*sum(charging)
                elif e_error<0 and sum(discharging)>0: #if discharging and less demand than anticipated reduce discharging
                    r = 1 - max([0,(sum(discharging) + e_error)])/sum(discharging)
                    e_error += r*sum(discharging)

                for i,k in enumerate(i_storage):
                    if e_error>0:
                        power = -r*charging[i]*gen[k]['stor']['charge_eff'] #(negative) amount less power devoted to charging
                    else:
                        power = r*discharging[i]/gen[k]['stor']['disch_eff'] #amount less power from discharging
                    if subnet['abbreviation'] not in gen[i]['output'] and acdc and net == 'electrical':
                        power = power/dc2ac
                        solution['dispatch'][acdc_i[0]][1] += power
                        solution['generator_state'][gen[acdc_i[0]]['name']][0] += power
                    elif subnet['abbreviation'] not in gen[i]['output'] and acdc and net == 'direct_current':
                        power = power/ac2dc
                        solution['dispatch'][acdc_i[0]][1] -= power
                        solution['generator_state'][gen[acdc_i[0]]['name']][0] -= power
                    solution['dispatch'][k][1] += power*dt
                    solution['generator_state'][gen[k]['name']][0] -= power
                    solution['storage_state'][gen[k]['name']][0] += power*dt

                #Utility makes up the difference
                i = i_utility[0]
                if subnet['abbreviation'] in gen[i]['output']:
                    solution['dispatch'][i][1] += e_error
                elif acdc and net == 'electrical':
                    solution['dispatch'][i][1] += e_error/dc2ac
                    solution['dispatch'][acdc_i[0]][1] -= e_error/dc2ac
                elif acdc and net == 'direct_current':
                    solution['dispatch'][i][1] += e_error/ac2dc
                    solution['dispatch'][acdc_i[0]][1] += e_error/ac2dc
                solution['generator_state'][gen[i]['name']][0] = solution['dispatch'][i][1]
        if 'buildings' in subnet and net in ['district_heat','district_cool']:
            gs_now = {}
            for k in solution['generator_state']:
                gs_now[k] = solution['generator_state'][k][0]
            production = component_output(gen,subnet,gs_now)
            thermal = distribute_heat_cool(names,subnet,production,demand,request,marginal,capacity)
    return thermal


def distribute_heat_cool(names,subnet,production,demand,request,marginal,capacity):
    ''' Find the thermal energy available for buildings at each node in the thermal network
    Then evenly distrubute amongs buildings at that node to minimize percent deviation from request
    Requires a optimization if there is a thermal network with line losses 
    output: thermal a list of the thermal energy provided to each building'''
    nn = subnet['nodes']
    if len(nn) == 1: #can handle multiple buildings at a single thermal node
        net_thermal = sum(production)
        if nn[0] in list(demand.keys()):
            net_thermal -= sum(demand[nn[0]]['demand'])
        net_request = sum(request['nominal'])
        thermal = [j*(net_thermal/net_request)*1000 for j in request['nominal']]
    else:
        thermal = dispatch_network(names,subnet,production,demand,request,marginal,capacity,False)
    return thermal


import numpy as np
import copy

def lin_load_increase(T_eff,load,hvac_elec):
    # s = sorted(range(len(T_eff)), key=lambda k: T_eff[k])
    # n = len(s)
    # T_eff = [T_eff[s[i]] for i in range(n)]
    # load = [load[s[i]] for i in range(n)]
    # hvac_elec = [hvac_elec[s[i]] for i in range(n)]

    #interpolate to find when load starts increasing
    T = [T_eff[0],T_eff[-1]]
    UA = [0,0]
    hvac2elec = 0
    min_load = min(load)
    if not all([load[i] == load[0] for i in range(len(load))]):
        i = 0
        while i<len(T_eff)-1 and load[i]>min_load:
            i+=1
        i_min = copy.copy(i)
        while i<len(T_eff)-1 and load[i]<=min_load:
            i+=1
        i_max = copy.copy(i)

        ##Parameters for increase in heating/cooling with decrease in temperature
        if i_min == 1:
            UA[0] = (load[0] - min_load)/(T_eff[1] - T_eff[0])
            T[0] = T_eff[1] + min_load/UA[0] #intercept with given slope
        elif i_min!=0:
            p = np.polyfit(T_eff[:i_min],load[:i_min],1)
            if p[0]<0:
                UA[0] = float(-p[0])
                T[0] = -float(p[1])/float(p[0]) #intercept with given slope

        #Parameters for increase in heating/cooling with increase in temperature
        if i_max == len(T_eff)-2:
            UA[1] = (load[-1] - min_load)/(T_eff[-1] - T_eff[i_max])
            T[1] = T_eff[-1] - min_load/UA[1] #intercept with given slope
        elif i_max!=len(T_eff)-1:
            p = np.polyfit(T_eff[i_max:],load[i_max:],1)
            if p[0]>0:
                UA[1] = float(p[0])
                T[1] = -float(p[1])/float(p[0]) #intercept with min_load
    #Average slope of electric load vs. heating/cooling
        sindex = sorted(range(len(load)), key=lambda k: load[k])
        sort_load = [load[i] for i in sindex]
        sort_elec = [hvac_elec[i] for i in sindex]
        q = np.polyfit(sort_load,sort_elec,1)
        if q[0]>1e-6:
            hvac2elec = float(q[0])

    return T,UA,min_load,hvac2elec


'''
Run single step forward of multi-zone building simulation
'''

from eagers.building.basic.leward_calc import leward_calc
from eagers.building.basic.wind_speed_calc import wind_speed_calc
from eagers.building.basic.thermostat_humidistat import thermostat_humidistat
from eagers.building.basic.area_weighted_temperature import area_weighted_temperature
from eagers.building.process.load_sched import load_sched
from eagers.building.process.reset_loop_param import reset_loop_param
from eagers.building.process.update_model_states_implicit import update_model_states_implicit
from eagers.building.process.zone_loads import zone_loads
from eagers.building.process.ideal_hvac import ideal_hvac
from eagers.building.process.sim_air_loops import sim_air_loops
from eagers.building.process.net_water_loop_supply import net_water_loop_supply
from eagers.building.components.unitary_systems import unitary_systems
from eagers.building.components.air_terminals import air_terminals

def buildings_step(building,weather,T,w,air_nodes,plant_nodes,e_use,loads,gains,heating,cooling,T_optimal,T_nominal,schedules,frost,dt,date):
    nz = len(building.zones['name'])
    cat = ['net_elec','heat_elec','heat_gas','cool_elec','fan_elec','other_fans','water_gas','water_elec','tower_elec','pumps']
    leward = leward_calc(weather['wdir'],building.surfaces['exterior']['normal'])
    wind_speed = wind_speed_calc(weather['wspd'],building.ctf['s_height'],building.site['terrain'])
    schedules = load_sched(building.schedule,building.holidays,[date],None,False)
        
    ###Load values specific to this moment in time
    T_set,w_set = thermostat_humidistat(building,schedules,T['zone']) ## HVAC temperature Schedules
    _,_,occupancy,mixing,infiltration,S,frost,T_mains,water_heat = zone_loads(building,date,schedules,weather,T['zone'],w['zone'],dt,frost)

    ## extrapolate 'optimal' building temp to all zones
    T_set_mod = optim_T_set(T_set,T_optimal,T_nominal,building)
    ###Determine HVAC setpoints
    air_nodes,central_flow,direct_flow,mixed_air_loop,direct_return_air,plenum,T,w,fan_avail = ideal_hvac(building,schedules,leward,wind_speed,S,air_nodes,weather['t_dryb'],mixing,infiltration,gains,T,w,T_set_mod,w_set,occupancy,dt,False)
    j = 0
    while j <=2: #iterate to ensure water loops reach correct temperature
        j+=1
        for x in cat:
            e_use[x] = 0
        plant_nodes['demand_temperature'] = [plant_nodes['demand_temperature'][building.plant_demand_nodes['inlet_node'][building.plant_demand_nodes['loop'][i]]] for i in range(len(building.plant_demand_nodes['loop']))]
        supply_flow,e_use,plant_nodes = sim_air_loops(building,air_nodes,mixed_air_loop,central_flow,w['loop'],weather['t_dryb'],w['air'],plant_nodes,e_use,fan_avail)
        supply_flow,e_use,plant_nodes = air_terminals(building,air_nodes,supply_flow,e_use,plant_nodes,T['central'])
        supply_flow,plant_nodes,e_use = unitary_systems(building,air_nodes,supply_flow,direct_flow,direct_return_air,T['direct'],weather['t_dryb'],w['air'],fan_avail,plant_nodes,e_use)#zone equipment not part of central air loops
        e_use,plant_nodes,loop_load = net_water_loop_supply(building,schedules,dt,plant_nodes,e_use,T_mains,weather['t_dryb'],w['air'],heating,cooling)

    e_use['water_gas'] += water_heat/building.impact_factor['steam_efficiency'][0] #unconnected water heat is considered purchased heat
    nonHVAC_electric = sum([loads['lighting_internal'][z]*loads['multiplier'][z] for z in range(nz)]) + sum([loads['plug_load'][z]*loads['multiplier'][z] for z in range(nz)]) + loads['exterior']['lighting'] + loads['exterior']['equipment'] + sum(loads['case']['electric']) + sum(loads['rack']['electric'])
    e_use['net_elec'] = nonHVAC_electric + e_use['pumps'] + e_use['heat_elec'] + e_use['fan_elec'] + e_use['water_elec'] + e_use['tower_elec']

    ## simulate a time step
    T['zone'],T['surf'],w['zone'],e_dict = update_model_states_implicit(building,T,S,wind_speed,leward,gains,dt,w,mixing,plenum,infiltration,supply_flow)
    humidity = w['zone']
    return T,humidity,frost,e_use,loop_load,air_nodes,plant_nodes

def optim_T_set(T_set,T_optim,T_nom,building):
    T_set_mod = {}
    T_set_mod['no_hvac'] = [j for j in T_set['no_hvac']]
    T_set_mod['heat'] = [j for j in T_set['heat']]
    T_set_mod['cool'] = [j for j in T_set['cool']]
    T_set_mod['dual_sp'] = False
    if T_set['dual_sp']:
        T_set_mod['dual_sp'] = True
    nz = len(T_set['cool']) 
    error = 1
    T_nom1 = area_weighted_temperature(building,T_nom)
    delta =  (T_optim-T_nom1)
    while abs(error)>1e-3:
        T_set_mod['cool'] = [T_nom[z] + delta for z in range(nz)]
        T = area_weighted_temperature(building,T_set_mod['cool'])
        error = T - T_optim
        delta += error/10
    T_set_mod['heat'] = [j for j in T_set_mod['cool']]
    return T_set_mod


"""Defines TimeSeriesDataSet class, which establishes common behavior
for time series data sets, and its subclasses.

Classes:
TimeSeriesDataSet
NodeDataTimeSeries
WeatherTimeSeries
"""

import calendar
from contextlib import contextmanager
import datetime as dt
from functools import wraps

import numpy as np
from numpy.lib import recfunctions as rfn
import tables as tb

from eagers.class_definition.data_constructed import DataConstructed
from eagers.config.dataset import DATASET_TYPES
from eagers.config.path_spec import HDF5_SUFFIX, USER_DIR_HDF5_DATASETS
from eagers.basic.file_handling import ensure_suffix
from eagers.basic.hdf5 import h5file_context, DatetimeFloatConverter as DFC


class TimeSeriesDataSet(DataConstructed):
    """Holds time series data set file information for on-demand access
    to its data. The two most useful methods are read() and
    read_timestamp_range().
    """

    def __init__(self, filename, dset_type):
        """TimeSeriesDataSet objects should not be instantiated
        directly.
        """
        assert filename is not None, "filename cannot be None."
        assert dset_type in DATASET_TYPES, \
            f"{dset_type!r} not an acceptable data set type."
        self.filename = filename
        self.dset_type = dset_type

    def __repr__(self):
        return f"<{type(self).__name__}(filename={self.filename!r})>"

    @property
    def filepath(self):
        return (USER_DIR_HDF5_DATASETS
            / self.dset_type
            / ensure_suffix(self.filename, HDF5_SUFFIX))

    @property
    def tablenames(self):
        with self.file_context() as h5f:
            root = h5f.root
            return list(c for c in root._v_children
                if isinstance(getattr(root, c), tb.table.Table))

    def colnames(self, tablename):
        """Reads the column names from the specified HDF5 table."""
        with self.table_context(tablename) as table:
            return table.colnames

    @classmethod
    def from_file(cls, name, *args, **kwargs):
        """Provides an implementation for DataConstructed.from_file().
        """
        return cls(name, *args, **kwargs)

    @contextmanager
    def file_context(self):
        """Exposes an HDF5 file while open."""
        with h5file_context(self.filepath, mode='r') as h5f:
            yield h5f

    @contextmanager
    def table_context(self, name):
        """Exposes an HDF5 table while open."""
        with self.file_context() as h5f:
            table = getattr(h5f.root, name)
            yield table

    @wraps(tb.Table.read)
    def read(self, tablename, *args, **kwargs):
        """Reference PyTables.Table.read() for documentation. The only
        difference is that this function also accepts a tablename
        argument before any others.
        """
        with self.table_context(tablename) as table:
            return table.read(*args, **kwargs)

    @classmethod
    def _read_timestamp_range_from_table(
        cls,
        table,
        start=None,
        stop=None,
        *,
        field=None,
        inclusive=True,
        return_indices=False,
    ):
        """Basic version of read_timestamp_range where a Table object is
        provided. This exists for internal use, to avoid having to close
        and re-open the file in some cases.
        """
        col = 'timestamp'
        start, stop = cls._handle_start_stop_args(
            table, start, stop, typecast="float"
        )
        # Check input integrity.
        assert start <= stop, "Datetime range must be in chronological order."
        # Create query.
        upper_inequality = '<=' if inclusive else '<'
        query_condition = (
            f"({col} >= {start}) "
            f"& ({col} {upper_inequality} {stop})"
        )
        # Since no variables from the current namespace are used, set
        # condvars to [].
        if return_indices:
            # 2 return values.
            indices = table.get_where_list(query_condition, condvars=[])
            return (
                table.read(indices[0], indices[-1] + 1, field=field),
                indices,
            )
        else:
            # 1 return value.
            return table.read_where(query_condition, condvars=[], field=field)

    @classmethod
    def _handle_start_stop_args(cls, table, start, stop, typecast=None):
        """Returns given start and stop dates, ensuring they are not
        None.

        Positional arguments:
        table - (PyTables.Table) Table to read from.
        start - (datetime, float, or None) Start date.
        stop - (datetime, float, or None) Stop date.

        Keyword arguments:
        typecast - (str) (Default: None) Type to cast to.
        """
        # Handle unspecified bounds.
        if start is None:
            start = table.read(0, 1, field="timestamp")[0]
        if stop is None:
            stop = table.read(-1, field="timestamp")[0]
        # Handle type casting.
        if typecast == "datetime":
            if not isinstance(start, dt.datetime):
                start = DFC.f2d_sgl2sgl(start)
            if not isinstance(stop, dt.datetime):
                stop = DFC.f2d_sgl2sgl(stop)
        elif typecast == "float":
            if not isinstance(start, float):
                start = DFC.d2f_sgl2sgl(start)
            if not isinstance(stop, float):
                stop = DFC.d2f_sgl2sgl(stop)
        return start, stop

    def read_timestamp_range(
        self,
        tablename,
        start=None,
        stop=None,
        *,
        field=None,
        inclusive=True,
        return_indices=False,
    ) -> np.recarray:
        """Return NumPy record array of the rows corresponding to the
        specified timestamp bounds (inclusive). If the given bounds
        don't exactly match time stamps in the file, the closest values
        within the given bounds are returned.

        Positional arguments:
        tablename - (str) The name of the Table to read from.

        Keyword arguments:
        start - (datetime or float) (Default: None) Lower bound of
            timestamp range. If None, read from the beginning of the
            data set.
        stop - (datetime or float) (Default: None) Upper bound of
            timestamp range. If None, read to the end of the data set.
            If this value is beyond the end of the data set, this method
            will read to the end.
        field - (str) (Default: None) Column name of data to be
            returned.
        inclusive - (bool) (Default: True) Whether upper bound
            inequality is inclusive (<= as opposed to <).
        return_indices - (bool) (Default: False) If set to True, the
            indices of the returned rows will also be returned. When
            used in conjunction with read(), this can be useful for
            getting points on either side of timestamp ranges.
        """
        with self.table_context(tablename) as table:
            return self._read_timestamp_range_from_table(
                table,
                start,
                stop,
                field=field,
                inclusive=inclusive,
                return_indices=return_indices,
            )


class NodeDataTimeSeries(TimeSeriesDataSet):
    def __init__(self, filename):
        super().__init__(filename, 'nodedata')

    def start_dates(self, as_datetime=False):
        """Returns an array of the available start dates for this data
        set.

        Keyword arguments:
        as_datetime - (bool) (Default: False) Whether to return a list
            of datetime objects instead of an array of floats.
        """
        with self.file_context() as h5f:
            floats = getattr(h5f.root, 'available_start_dates').read()
            return DFC.f2d_arr2lst(floats) if as_datetime else floats


class WeatherTimeSeries(TimeSeriesDataSet):
    # tmy: Typical Meteorological Year. Timestamps only for one non-leap
    # year.
    # continuous: Continuous data. Timestamps possibly spanning multiple
    # years.
    _designations = ('tmy', 'continuous')

    def __init__(self, filename):
        super().__init__(filename, 'weather')
        # Check designations.
        self._des = {}
        for tablename in self.tablenames:
            with self.table_context(tablename) as table:
                des = table.attrs.designation
            if des not in self._designations:
                raise ValueError(
                    f"{des!r} in {filename!r} not a recognized designation.\n"
                    f"Acceptable designations: {', '.join(self._designations)}"
                )
            self._des[tablename] = des

    @wraps(TimeSeriesDataSet.read_timestamp_range)
    def read_timestamp_range(
        self, tablename, start=None, stop=None, *args, **kwargs
    ):
        designation = self._des[tablename]
        if designation == 'continuous':
            # Use the default method.
            return super().read_timestamp_range(
                tablename, start, stop, *args, **kwargs
            )
        elif designation == 'tmy':
            with self.table_context(tablename) as table:
                return self._tmy_read_timestamp_range(
                    table, start, stop, *args, **kwargs
                )

    def table_designation(self, tablename):
        """The weather file designation of the table with the given
        name.
        """
        try:
            return self._des[tablename]
        except KeyError:
            raise ValueError(
                f"Couldn't find a designation for table {tablename!r}"
            )

    def table_designation(self, tablename):
        """The weather file designation of the table with the given
        name.
        """
        try:
            return self._des[tablename]
        except KeyError:
            raise ValueError(
                f"Couldn't find a designation for table {tablename!r}"
            )

    def table_designation(self, tablename):
        """The weather file designation of the table with the given
        name.
        """
        try:
            return self._des[tablename]
        except KeyError:
            raise ValueError(
                f"Couldn't find a designation for table {tablename!r}"
            )

    def _tmy_read_timestamp_range(
        self,
        table,
        start,
        stop,
        *args,
        return_indices=False,
        **kwargs,
    ):
        """Core logic for read_timestamp_range specific to TMY-type
        weather files and where a table is provided. This exists for
        internal use, to avoid having to close and re-open the file in
        some cases.

        TMY weather files assume a non-leap year. When the requested
        range includes February 29th, the missing data must be made up
        for. In this case, Feb. 28 data is used in place of Feb. 29.

        Positional and keyword arguments are the same as for
        read_timestamp_range, except for this additional keyword
        argument:
        impose_year - (bool) (Default: True) Indicates whether the
            requested range should be returned with the years of the
            requested datetimes imposed on the data read from the TMY
            file. Otherwise, the timestamps for the returned data will
            all have the year of the TMY file. E.g. if the range from
            2011-4-23 to 2012-7-4 is requested from a file with the TMY
            year 2017 and impose_year is False, the data returned will
            have timestamps from 2017-4-23 to 2017-12-31, then 2017-1-1
            to 2017-7-4.
        """
        start, stop = self._handle_start_stop_args(
            table, start, stop, typecast="datetime"
        )
        # Check for Feb 29 within requested range.  If found, will need
        # to break at occurrences of both Feb 29 and new year.  After
        # initializing an array, append sections of the range between
        # breaks until end of requested range.
        tmy_yr = DFC.f2d_sgl2sgl(table.read(0, 1, field='timestamp')[0]).year
        range_sections, years = self._tmy_year_sections(start, stop, tmy_yr)
        arrays = []
        indices = []
        for i, year in enumerate(years[:-1]):
            start, stop = range_sections[i]
            # Use inclusive=False.  All tuples except the last one
            # indicate non-inclusive sections.
            self._tmy_read_year_section(
                table,
                start,
                stop,
                year,
                arrays,
                indices,
                *args,
                inclusive=False,
                return_indices=return_indices,
                **kwargs,
            )
        for i, year in enumerate(years[-1:]):
            start, stop = range_sections[-1]
            # Final tuple is inclusive; use inclusive=True.
            self._tmy_read_year_section(
                table,
                start,
                stop,
                year,
                arrays,
                indices,
                *args,
                inclusive=True,
                return_indices=return_indices,
                **kwargs,
            )
        data = rfn.stack_arrays(arrays, usemask=False, asrecarray=True)
        if return_indices:
            return data, np.hstack(indices)
        else:
            return data

    @staticmethod
    def _tmy_year_sections(start, stop, tmy_year):
        """Return a list of two-tuples representing start and end times
        for sections of the given range to be read from the TMY data
        file and a list of the requested years corresponding 1-to-1 with
        each section. Section breaks occur at new years and the start
        and end of leap days. Because breaks occur at new years, each
        section will correspond to only one year. Since TMY data files
        contain data for non-leap years, Feb. 28 is used in place of
        Feb. 29.
        
        End times are non-inclusive to avoid needing to know the
        smallest time increment. E.g. whether the entire day of February
        28th, 2017 (a non-leap year) is denoted as
        (2017/2/28 00:00, 2017/2/28 23:00) or
        (2017/2/28 00:00, 2017/2/28 23:59) depends on whether the
        smallest time increment is 1 hour or 1 minute. Using the non-
        inclusive notation (2017/2/28 00:00, 2017/3/1 00:00) to denote
        the section of time starting at the beginning of 2/28 and ending
        just before the beginning of 3/1 eliminates ambiguity.
        Positional arguments:
        start - (datetime) Start of range.
        stop - (datetime) End of range.
        tmy_year - (int) Year of the TMY data set.
        """
        # Define boolean functions for determining specific dates.
        def is_feb29(x):
            try:
                return x.date() == dt.date(x.year, 2, 29)
            except ValueError:
                # Not a leap year; can't make a date for 2/29.
                return False

        def is_new_year(x):
            return x == dt.datetime(x.year, 1, 1)

        # Find section breaks.  These occur at the start of a new year,
        # the start and end of Feb 29, and the start and stop dates.
        # The requested years are tracked at the same time.  This
        # procedure will result in breaks being one element longer than
        # req_yrs, since each year will correspond to the sections of
        # time bounded by pairing each element of breaks with the next
        # one.
        # The inequality used for the while condition is inclusive.
        # When the cursor ends on the stop date, one final append of the
        # cursor and year is required.
        breaks = []
        req_yrs = []
        cursor = start
        while cursor <= stop:
            breaks.append(cursor)
            req_yrs.append(cursor.year)
            if calendar.isleap(cursor.year):
                cursor_date = cursor.date()
                feb_29 = dt.date(cursor.year, 2, 29)
                if cursor_date < feb_29:
                    cursor = dt.datetime(cursor.year, 2, 29)
                elif cursor_date == feb_29:
                    cursor = dt.datetime(cursor.year, 3, 1)
                else:
                    cursor = dt.datetime(cursor.year + 1, 1, 1)
            else:
                cursor = dt.datetime(cursor.year + 1, 1, 1)
        breaks.append(stop)

        # Convert section breaks to sections.  For 2/29, read 2/28.
        # Modify year to ensure sections fall within the TMY year, i.e.
        # so the requested data is actually found in the TMY file.
        sections = []
        for b1, b2 in zip(breaks[:-1], breaks[1:]):
            x1 = b1.replace(day=28) if is_feb29(b1) else b1
            x1 = x1.replace(year=tmy_year)
            if is_feb29(b2):
                if b2 > dt.datetime(b2.year, 2, 29):
                    # Upper bound is somewhere in the middle of 2/29.
                    x2 = b2.replace(day=28)
                else:
                    # Upper bound is exactly the start of 2/29.
                    if is_feb29(b1):
                        # An occurrence of 2/29 in both bounds where b2
                        # is on hour zero can only mean that one data
                        # point on 2/29 is requested.
                        x2 = b2.replace(day=28)
                    else:
                        # Read up to the end of 2/28 in the TMY file.
                        x2 = b2.replace(month=3, day=1)
                x2 = x2.replace(year=tmy_year)
            # An occurrence of the new year where both bounds have the
            # same year can only mean that one final data point on Jan.
            # 1st is requested.
            elif is_new_year(b2) and b2.year != b1.year:
                x2 = b2.replace(year=tmy_year+1)
            else:
                x2 = b2.replace(year=tmy_year)
            sections.append((x1, x2))
        return sections, req_yrs

    def _tmy_read_year_section(
        self,
        table,
        start,
        stop,
        year,
        arrays,
        indices,
        *args,
        return_indices=False,
        impose_year=True,
        **kwargs,
    ):
        """Read a section from the TMY file up to the end of the year.
        """
        result = self._read_timestamp_range_from_table(
            table,
            start,
            stop,
            *args,
            return_indices=return_indices,
            **kwargs,
        )
        if return_indices:
            result, idx = result
            indices.append(idx)
        if (
            impose_year
            and result.dtype.names is not None
            and 'timestamp' in result.dtype.names
        ):
            # Replace the TMY file's year with the requested one.
            result['timestamp'] = DFC.d2f_lst2arr(
                [
                    x.replace(year=year)
                    for x in DFC.f2d_arr2lst(result['timestamp'])
                ]
            )
        arrays.append(result)



def window_solar_transmittance(windows,surfaces,vect_2_sun,norm_irrad,horz_irrad,visible,long_wave):
    # outputs:
    # S is energy hitting surface of windows, 
    n_win = len(windows['name'])
    window_gain = [0 for i in range(len(surfaces['name']))]
    S = [[0 for i in range(n_win)] for j in range(4)]

    cos_phi_sun = [max(0,sum([vect_2_sun[j]*windows['normal'][i][j] for j in range(3)])) for i in range(n_win)]    
    ## transmit/reflect for glazing and simple glazing 
    transmit_angle_mod = [windows['transmittance'][i][0]*cos_phi_sun[i]**4 + windows['transmittance'][i][1]*cos_phi_sun[i]**3 + windows['transmittance'][i][2]*cos_phi_sun[i]**2 + windows['transmittance'][i][3]*cos_phi_sun[i] + windows['transmittance'][i][4] for i in range(n_win)]    
    reflect_angle_mod = [windows['reflectance'][i][0]*cos_phi_sun[i]**4 + windows['reflectance'][i][1]*cos_phi_sun[i]**3 + windows['reflectance'][i][2]*cos_phi_sun[i]**2 + windows['reflectance'][i][3]*cos_phi_sun[i] + windows['reflectance'][i][4] for i in range(n_win)]    
    reflect = []
    transmit = []
    visible_absorptance = []
    for i in range(n_win):
        transmit.append([windows['normal_transmittance'][i][0]*transmit_angle_mod[i],
                         windows['normal_transmittance'][i][0]*transmit_angle_mod[i],
                         windows['normal_transmittance'][i][1]*transmit_angle_mod[i],
                         windows['normal_transmittance'][i][1]*transmit_angle_mod[i]])
        reflect.append([windows['normal_reflectance'][i][j]*reflect_angle_mod[i] for j in range(4)])
        visible_absorptance.append([max(0,1 - transmit[i][j] - reflect[i][j]) for j in range(4)])##solve equation on page 290 for absorptance, for single layer it is what is not transmitted or reflected
    diffuse_absorptance = [[visible_absorptance[i][j] for j in range(4)] for i in range(n_win)] #cant find any mention describing how to get this paramater for simple window model

    for i in range(n_win):
        if windows['type'][i]=='SimpleGlazingSystem':
                S[0][i] = .5*(norm_irrad*cos_phi_sun[i]*visible_absorptance[i][0] + horz_irrad*diffuse_absorptance[i][0] + visible[i]*(1 - windows['normal_transmittance'][i][0] - windows['normal_reflectance'][i][0]))
                S[1][i] = S[0][i] + windows['emittance'][i][1]*long_wave[i] #add interior radiation to inside of 1st pane for simpleglazing
        else:
                S[0][i] = .5*(norm_irrad*cos_phi_sun[i]*visible_absorptance[i][0] + horz_irrad*diffuse_absorptance[i][0] + visible[i]*windows['normal_transmittance'][i][1]*(1 - windows['normal_transmittance'][i][0] - windows['normal_reflectance'][i][0]))
                S[1][i] = S[0][i]
                S[2][i] = .5*(norm_irrad*cos_phi_sun[i]*transmit[i][0]*visible_absorptance[i][2] + horz_irrad*windows['normal_transmittance'][i][0]*diffuse_absorptance[i][2] + visible[i]*(1 - windows['normal_transmittance'][i][1] + windows['normal_reflectance'][i][3]))
                S[3][i] = S[2][i] + windows['emittance'][i][1]*long_wave[i] #add interior radiation to inside of 1st pane for simpleglazing
        
        #solar distribution (page 212 of reference) 
    if norm_irrad>0 or horz_irrad>0:
        T_incident = []
        for i in range(n_win):
            if windows['type'][i]=='Glazing':
                T_incident.append(transmit[i][0]*transmit[i][2]/(1-reflect[i][2]*reflect[i][1]))
            else:
                T_incident.append(transmit[i][0]) #net transmittance factor for direct sunlight    

        for i in range(n_win):
            z = windows['zone_index'][i]
            surf = [x for x in range(len(surfaces['name'])) if surfaces['zone_index'][x]==z and surfaces['type'][x]!='InternalMass'] #no solar gain onto interior furnishings
            z_surf_area = sum([surfaces['area'][j] for j in surf])
            f = [x for x in surf if  surfaces['surf_type'][x] == 'floor']
            tot_area = sum([surfaces['area'][x] for x in f]) 
            f_absorb = [surfaces['absorptance_visible'][x][0]*surfaces['area'][x]/tot_area for x in f] #for things like corridors with multiple 'floor' surfaces
            normal2zone = norm_irrad*cos_phi_sun[i]*T_incident[i]*windows['area'][i]
            horizontal_irrad = horz_irrad*windows['normal_transmittance'][i][0]*windows['area'][i] + normal2zone*(1-sum(f_absorb))
            for j in range(len(f)):
                window_gain[f[j]] += normal2zone*f_absorb[j]
            for j in surf:
                window_gain[j] += horizontal_irrad*surfaces['area'][j]/z_surf_area #distribute diffuse sunlight to surfaces in zone

    return window_gain,S    